/// ASM85 - Everett Kaser - July 2020
/// Linux port and enhancements by Philippe Bergheaud

// DEVTIME sets default command line values when debugging in the development environment.
// Set it to 0 before release.
#define	DEVTIME	0

/// NOTE! This assembler is CASE SENSTIVE in regard to LABELS!
/// The label "HELLO" is NOT the same as "hello" or "Hello" or any other combinations!

/********************************************************
 TO-DO LIST:
	fix ABS 16 and ABS 32 for 85, ABS BPGM for 87
	maybe add support for bin constants
 ********************************************************/

/*********************  Header Files  *********************/
#if defined __linux__ || defined __APPLE__
	#include <stdlib.h>
	#include <strings.h>
	#include <unistd.h>
	#define BOOL unsigned int
	#define BYTE unsigned char
	#define DWORD unsigned int
	#define FALSE 0
	#define _lseek(a, b, c) lseek(a, b, c)
	#define max(a,b) (((a) > (b)) ? (a) : (b))
	#define min(a,b) (((a) < (b)) ? (a) : (b))
	#define _O_BINARY 0
	#define _O_CREAT O_CREAT
	#define _O_RDONLY O_RDONLY
	#define _O_RDWR O_RDWR
	#define _O_TRUNC O_TRUNC
	#define _S_IREAD S_IREAD
	#define _S_IWRITE S_IWRITE
	#define strnicmp(a, b, c) strncasecmp(a, b, c)
	#define SYSTEMTIME struct tm *
	#define TRUE 1
#else
	#include <windows.h>
	#include <io.h>
#endif

#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <string.h>
#include <sys/stat.h>
#include <ctype.h>

/*
0.01 17 JUL 2020 - first prototype version
1.00 6 AUG 2020
1.01 7 AUG 2020
	FIXED: LDM R*,=125,77 would generate a "ARG LIST TOO LONG" error
1.02 10 AUG 2020
	ADDED: OCT opcode, equivalent to BYT
1.03 13 AUG 2020
	ADDED: full-line-comment meta-controls, where the line starts with "!(xxx)", where "xxx" is the meta-control and all other assemblers will see it as just a full-line-comment. Anything after the "!(xxx)" is treated as a comment
		!(COM+) and !(COM-) meta-controls to start/end (possibly nested) commented blocks of lines
		!(OK) meta-control to suppress any WARNING on the next line of actual code
		!(UNREF+) and !(UNREF-) to turn on and off suppression of "unreferenced label" warnings for all labels defined while it's ON(+)
	REMOVED: pseudo-opcodes "OKAY" and "UNREF".
1.04 17 AUG 2020
	FIXED: if src file didn't exist, empty .lst and .bin files were still created.
	FIXED: !(UNREF) wasn't working right.
	ADDED: if a HED has been done, auto-add a HED "LABEL REFERENCE" before dumping label table
1.05 19 AUG 2020
	FIXED: ARP/DRP wasn't being invalidated after PAD
	ADDED: saving of 'C' & 'D' format specifiers on EQU & DAD definitions, and then dumping them that way in the LABEL REFERENCE rather than as octal.
1.06 21 AUG 2020
	ADDED: filename and line# on error output (involved adding incSTK... and fake " EINC\r\n" line at end of INC'd file and linenumBase, to keep things straight);
	FIXED: HED wasn't showning the proper source filename after end of an INC'd file.
	FIXED: -c help had wrong sense, it DISABLES cleaning of source code, not ENABLES it.
1.07 28 AUG 2020
	FIXED: "ASC 5,abcdefghijk" would give an error, because it only chewed up the first 5 chars of the unquoted string. CASM ignored the rest, so we do, too.
	ADDED: DEC opcode which takes a decimal value of up to 16 digits and generates as many BCD digit bytes equivalent to it in the code. This was used a little in the original system ROM source code.
	FIXED: DRP instruction was ignored, when the specified register matched the existing DPR (same for ARP). Now, it FORCES the DRP/ARP regardless, as CASM did.
	FIXED: LD/ST/AD/CM R57,=label where 'label' was a DAD was pushing two bytes and giving a warning (because it wasn't an EQU).
1.08 12 SEP 2020
	CHANGED: combined aCmAdSuAn and aLDST, so all those opcodes use the same aLDST routine with a few additional checks. Now, any bug fixes only have to be made in one place. :-)
	ADDED: Table of Contents (dump of HEDs and page#s on which they start) and !(TITLE) meta-command comment (lines of !(TITLE) text output above Table of Contents)
	CHANGED: LF var to FF, as it's a FORM Feed, not a LINE Feed. :-)
	ADDED: !(85), !(87), !(75) meta-comment commands to specify the MACH in the source code itself.
			It can still be specified on the command line, if preferred, but one way or the other, it can't be specified twice.
	CHANGED: Loading/processing of GLOBALS doesn't happen until MACH is set, which MUST happen before any code generation begins.
	FIXED: HED src filename for INC'd and LNK'd files could be too long and run into the date.
	ADDED: -s option to make HED headings SECTION headers rather than page headers, to only get spit when the HED statement is encountered, and no further HED blocks & page breaks thereafter until another HED is seen.
	CHANGED: TABs in source code get converted into spaces in .lst file, assuming TABs = 4 spaces, a TAB outputs spaces until the column is at the next multiple of 4 columns.
	FIXED: spaces were stripped from quoted strings in .LST file for ASC, ASP, INC, LNK opcodes.
	ADDED: Ability to define multiple labels with the same value via:
		NEWLABEL EQU OLDLABEL
		NEWLABEL DAD OLDLABEL
	FIXED: EOL comments, when 'cleaning' the code for listing, weren't being placed far enough right.
1.09 15 SEP 2020
	ADDED: possible !(cmd) meta-comment IN FIRST LINE ONLY!, to set command-line options IN the source code so they don't have to be typed on the command line or in a batch file
		NOTE: simple switches (s, c, w, etc) simply toggle the setting, so command-line can override what's in the source file, too.
			For the -85/-87/-75 switches, if the switch is on both the command-line and the source-first-line, those switches MUST match or an error will occur.
	ADDED: ability to use "'x'" (matched single-quotes with an ASCII char in between) in place of of a literal number.
	ADDED: ability to '~' in front of literal numbers and labels to do a bit-wise NOT (1's complement) on the argument, but not when being used as an address.
1.10 24 SEP 2020
	FIXED: ldm r40,=' ' would get the SPACE removed from between the ''s by CleanSrc() in the .LST file.
1.11 16 NOV 2020
	FIXED: I didn't understand the proper operation of the DEC pseudo-opcode. Upon discovery of further examples, it's now clear that it takes up to 16 decimal digits
	    and converts it to bytes with the corresponding BCD digits, and it generates ONLY enough bytes to hold the specified digits. So it could generate anywhere from 1 to 8 bytes of code.
1.12 26 NOV 2020
	ADDED: When 'mach' is set, "HP85", "HP87", "HP75" local labels are created as AIF type and set to 0 except for the one that matches 'mach'.
		This saves the assembly programmer from having to create and manage similar labels in their source code for conditional assembly.
	FIXED: errors occuring in INC'd files reported a line# one higher than what it should have been.
	CHANGED: by default, conditional assembly that's not assembled, including the surrounding AIF/ELS/EIF statements, is NOT shown in the .LST file.
	ADDED: -a switch to force ALL conditional assembly code to be shown in the .LST file, whether it's assembled or not.
1.13 16 FEB 2021
	CHANGED: HED lines are no longer shown in .LST file.
1.14 17 APR 2021
	CHANGED: All 'system' type errors (file access, invalid command line arguments, etc) return an exit() value of -1.
		Otherwise, for normal termination, return value is !!(ErrCnt + WarnCnt) for batch processing control.
	ADDED: to the .LST file and the standard output for ROM assemblies: "n ROM filler bytes added", for ease of knowing how much space is left in the ROM.
2.00 29 SEP 2023
	ADDED: changes from Philippe Bergheaud for Linux #if's and !(PUBLIC) meta-comment feature (see docs\manual.txt).
	CHANGED: Philippe's "#ifdef __linux__" statements to "#if defined __linux__ || defined __APPLE__" for linux AND Mac OS compiles.
2.01 6 Nov 2023
	FIXED: Absolute BPGMs weren't getting their base address set in PCB
	FIXED: relative BPGMs were getting xFFFF stored for base address in PCB
2.02 18 Nov 2023
	ADDED: BRK opcode (for opcode 336, which is a NO-OP). Spits warning.
2.03 24 Nov 2023
	FIXED: STM=LIT opcode wasn't included in the "check if !(OK) is acceptable" for missing arguments.
2.04  5 Dec 2023
	FIXED: GTO in non-absolute BPGM was still doing LDM instead of ADM
2.05 11 Dec 2023
	ADDED: GetLabel() sets OneByteLabel FALSE unless there's are {}'s around the label. aLDST multibyte then checks if OneByteLabel is set and only outputs the LOW byte value of the label.
		This allows this:
			MYROM#	EQU	13
			LDM R56,={MYROM#},370
		Without the {}'s around MYROM#, ASM85 would put 13 into R56 and 0 into R57, and then error when there was another value (370) in the list, because it had run out of registers.
	ADDED: -B85, -R85, -B87, -R87 command line switches to indicate BPGM or ROM and preset isROM, isROM85, isROM87, isBPGM, isBPGM85, isBPGM87 conditional assembly (AIF) flags (in addition to HP85 and HP87 flags)
	ADDED: if not already set by command line switches, isROM, isROM85, isROM87, isBPGM, isBPGM85, isBPGM87 are set by ABS and NAM statements.
		With these conditional assembly flags, a single source file can be constructed that can be built to generate either a binary program or ROM for either an 85  or 87 model via a single command-line switch.
	ADDED: ELIF <conditional label> to allowed a string of conditional assemblies: AIF <label>, ELIF <label, ELIF <label> .... [optional final ELS] EIF. Only ONE section of the string will assemble.
2.06 20 Dec 2023
	ADDED: leading line numbers (as you get when PLISTing source code from the Emulator) get "auto-stripped" by default. If a
		source doesn't HAVE leading line numbers, but DOES have labels that start with a digit (or some other unforeseen
		circumstance), the auto-stripping of line numbers can be disabled via the "-#" command line switch. Stripping can be
		selectively turned off and on using the "!(#KEEP)" and "!(#STRIP)" meta commands within source files.
	ADDED: "==========================>> DANGER, WILL ROBINSON!" after ERRORS/WARNINGS count messages at the end of the
		assembly when there ARE errors or warnings, to make it easier to spot problems when assembling a long list of
		source files in a batch file (such as CHKASM.BAT in the ROM source distribution).
2.07  1 Jan 2024
	ADDED: HP-75 ROM 1-byte checksum
	ADDED: Support for HEX constants (if leading character is a hex digit, then it MUST be preceded by a '0') that end with 'H' or 'h'
	ADDED: -h command line (or !(cmd)) switch to toggle address and object code output in the list file to be formatted in HEX or OCTAL. Default is HEX for 75, OCTAL for 85/87.
	ADDED: -n# command line (or !(cmd)) switch to set the "number of bytes of object code per line" in the list file, where the '#' must be a decimal # from 1 to 99.
	ADDED: !(#OBJ) meta-command that allows modifying the current "-n#" value dynamically throughout the source code listing.
		!(#OBJ) must be followed by a decimal # from 0 to 99. 1-99 sets that number of object bytes in the listing, going forward from that point.
		!(#OBJ) 0 restores the number of object bytes that was set either by default, by the -n# command line switch, or the !(cmd) -n# switch, going forward from that point.
		For example, you might want to set "!(#OBJ) 10" before the ASCIIS and ERMSG tables, and then at the end of the table restore the 'normal' setting with "!(#OBJ) 0".
	CHANGED: the optional lineWidth,pageHeight arguments on 'hed' were being reset to defaults any time they weren't included on a 'hed' statement.
		Now, they remain the same, i.e., once changed, they retain that change for the rest of the source file or until another 'hed' specifies new settings. The defaults are 95,60.
2.08	6 Jan 2024
	ADDED: check to GTO code in a relative BPGM, to spit a warning if the DCM flag state is not KNOWN to be in BIN mode. !(OK) before the GTO will suppress the warning if it's known to be in BIN mode.
	ADDED: after a label, JSB, or PAD, DRP and ARP are normally 'invalidated' so new ones will be generated. If you KNOW the DRP or ARP is a certain value after one of those situations in your code,
		you can now place a '^' character before the register# to indicate that you KNOW it's that value. That will set the DRP or ARP to that known value without generating a DRP or ARP instruction,
		so that that value can be used therein and thereafter without causes an unneeded DRP/ARP instruction to be generated.
	ADDED: '@' labels as targets for conditional jmp instructions (not legal for use with any other types of instructions). A label consists of a single '@' character at the start of a line,
		in place of a 'normal' label. It may be referenced in conditional jmp instructions as @- or @b/@B, and as @+ or @f/@F. A backwards reference targets the FIRST '@' label BEFORE the conditional
		jmp instruction (which MAY be on the SAME line as the conditional JMP, that's considered 'before'), whereas a forward reference targets the FIRST '@' label AFTER the conditional jmp.
		You can also use multiple repetitions of the -/b or +/f character to reference the SECOND, THIRD, etc. label BEFORE or AFTER the jmp instructions (such as JMP @--- or JNC @+fF), as long as
		all of the directional characters match in their 'direction'. NOTE: these kinds of labels can quickly become confusing, especially as code is edited over time, so it is recommended that
		only one or two levels be used. Great care must be observed beyond that. The '@' label mechanism is useful for not having to invent unique labels for trivial, 'local' branches. Sometimes
		labels are part of the 'documentation' of the code, though, so when '@' labels are used, comments are critical for documentation purposes.
	CHANGED: an optional '=' is now allowed in an indexed JSB instruction. BOTH "JSB X14,LABEL" and "JSB X14,=LABEL" are now allowed.
	CHANGED: on an indexed JSB, the 'X' is not enforced. If JSB isn't followed by an '=', then it's assumed to be an INDEXED JSB, and any normally allowable ARP format is accepted (e.g., R32, X32, ^32, 32).
	CHANGED: on an indexed LDM/STM, the 'X' is not enforced.
	ADDED: BINARY numeric constants (1011010111011B (or b)).
*/
char			*TitleRevStr = "ASM85 2.08 (2024 Jan 7)";

void AsmError(char *err);
void SpitHED();
BOOL GetLabel(char *label);

BYTE	ATlabels[8192];	// flags for the locations of '@' labels

char	FF=12;	// FF char
long	Okay;
int	Public;

#define	PATHLEN	512
char	srcFileName[PATHLEN], curFileName[PATHLEN], lstFileName[PATHLEN], gloFileName[PATHLEN], binFileName[PATHLEN], inclnkFileName[PATHLEN];
char	pubFileName[PATHLEN];

char	*EINC=" EINC\r\n";	// end-of-inc, pop-the-incSTK pseudo-op
char	incSTKname[10][PATHLEN];// for saving curFileName
int		incSTKlinenum[10][2];	// for saving linenum & linenumBase
int		incSTKcnt;

char	ampm[]="ampm";
char	Months[]="JanFebMarAprMayJunJulAugSepOctNovDec";

char	eolseq[]="\r\n";
int		eollen=2;
int		pass;

int		drp, arp, dcm;	// ==-1 if unkonwn
BYTE	code[128];
int		codelen;
int		linenum, linenumBase, InCommentBlk;
BOOL	StripLineNumbers=TRUE;

BOOL	NotDone;

long	UnRef;	// labels defined while UnRef==1 get listed at end of assembly if unreferenced, labels defined while UnRef==0 don't get listed even if unreferenced

long	AIFcnt;
unsigned short	AIFstack[256];
BYTE	isAIF, AIFstmt, isAIFdone;	// isAIFdone=0 until an AIF/ELSIF statement is true, then isAIFdone=1 until EIF. Both isAIF and isAIFdone get saved (combined) on AIFstack
BOOL	showAIF=FALSE;

long	GDADcnt, GEQUcnt, LCODcnt, LDADcnt, LEQUcnt, LAIFcnt;
int	PUBcnt[3];

DWORD	BSZcnt;

BOOL	ListOn=TRUE, isHEDline=FALSE;

char	ErrBuf[256], WarnBuf[256];
long	ErrCnt, WarnCnt;

struct page {
	int		pageN;		// page# where HED starts
	BYTE	*pHED;		// pointer to HED string
};
typedef struct page PAGE;

PAGE	TOC[512];		// Table-Of-Contents pointers
long	TOCcnt=0;
BYTE	*Titles[100];
long	TitlesCnt=0;

BOOL	secHED=FALSE;	// default to PAGE HEDs
BOOL	badS=FALSE;		// if HED appeared BEFORE !(s) on first pass
long	hedW1, hedW2, hedW3, lineW=95, lineH=60;
long	lineN, pageN;
char	HEDbuf[64], HEDsrc[PATHLEN], HEDobj[PATHLEN];
DWORD	finLen;

#define	MACH_85	0
#define	MACH_87 1
#define	MACH_75 2

long	mach=-1;		// default to HP-85 (0) unless -87 (1) or -75 (2) option is specified
long	baseAddr;		// -1 for BPGM, 0, 20000, 40000, 60000 for ROMs
long	nextAddr;
BOOL	DoWarnings=TRUE;	// if warn about non-errors
BOOL	seenNAM;
long	isROM=-1;		// 0 if -BPGM85 or -BPGM87 has been seen, 1 if -ROM85 || -ROM87 has been seen, else -1
long	isBPGM=-1;		// 1 if -BPGM85 or -BPGM87 has been seen, 0 if -ROM85 || -ROM87 has been seen, else -1

char	*srcBuf=NULL;
char	*pSRC;
long	srcBufLen;

int		lstFile, binFile, pubFile;

int		HexOut=0;
int		NumObj=-1;	// 1 less than the number of bytes of object code per line in .LST file
int		NumObjDef;	// default NumObj value, set on the command line or the !(cmd) line, used to restore it with "!(OBJ#) 0" after having "!(OBJ#) something"

typedef struct {
	char	*adrForm;	// formatting string for address
	char	*objForm;	// formatting string for object codes
	char	objNum;		// how many bytes of object code to display
	int		objW;		// how wide address/objCodes space is (distance from left column to start of LABELs)
} OUTFORM;

char	*pHexAdr="%04X ", *pOctAdr="%06o ", *pHexObj="%02X ", *pOctObj="%03o ";
OUTFORM	OutForm;
char	Spaces[128]="                                                                                                                                ";

char *globfile[]={
	"globals85.txt",
	"globals87.txt",
	"globals75.txt"
};

#define	DAD		0
#define	EQU		1
#define	LOC		2
#define	AIF		3

char	LabelType[]="DADEQULOCAIF";
char	PublicType[]="DADEQUDAD";

#define	BASEO	0
#define	BASEB	1
#define BASED	2
#define BASEC	3
#define	BASEH	4

struct label {
	char	type;
	char	numbase;	// BASEO, BASEB, BASED, BASEC, BASEH
	char	label[10];	// really only use 9 max, the others keep the structure a multiple of 4 in len
	int		line;		// linenum on which the label was first defined (set during FIRST PASS)
	int		refd;		// 0 if never referenced, 1 if referenced during the assembly (for reference table generation)
	int		unref;		// whether to list this (or not) at end of assembly if .refd ==0
	int		public;		// 0 if private, >0 if symbol is public
	char	*comment;
	DWORD	value;
	long	Lnext, Vnext;	//  index of next [] item when sorted by Label or Value (the [] is 'sorted' by the order in which the labels were encountered)
};
typedef struct label LABEL;

long	GfirstL, GfirstV, LfirstL, LfirstV;	// index of first label in Globals[] and Labels[] when sorted by Label or Value

LABEL	Globals[8192];
int		gloCnt;
LABEL	Locals[8192];
int		locCnt;

char	*gloBuf;	// ptr to malloc'd input GLOBALS file
char	*pGLO;		// working ptr into gloBuf
long	gloBufLen;	// len of gloBuf;

char	ScratBuf[8192];

char	*pLines[65536];	// ptrs to each line of the input file(s)
						// up to 8 pointers for EVERY byte of the ROM.  Should be enough.
int		linesCnt;

BOOL	CleanLST=TRUE;
BOOL	GetBNOT;		// whether the most recent GetNum or GetLabel was preceded by a ~ (ones complement) operator
#define BNOT(p) (GetBNOT?(~(p)):(p))
BOOL	OneByteLabel;	// if label is surrounded by {}'s, that means only use the LOW byte (treat is as a 0-255 numeric value)

int aCmAdSuAn(int iOPS);
int aAIF(int iOPS);
int aEIF(int iOPS);
int aELS(int iOPS);
int aELIF(int iOPS);
int aArpDrp(int iOPS);
int aAscAsp(int iOPS);
int aFin(int iOPS);
int aAbs(int iOPS);
int aValByt(int iOPS);
int aDef(int iOPS);
int aBsz(int iOPS);
int aNoArgs(int iOPS);
int aLoc(int iOPS);
int aJsb(int iOPS);
int aJmp(int iOPS);
int aLstUnl(int iOPS);
int aPushPop(int iOPS);
int aDR(int iOPS);
int aLDST(int iOPS);
int aGTO(int iOPS);
int aDRAR(int iOPS);
int aORG(int iOPS);
int aSetClr(int iOPS);
int aLNK(int iOPS);
int aHED(int iOPS);
int aNAM(int iOPS);
int aDEC(int iOPS);

typedef struct {
	DWORD	opname;		// 4 letters jambed together
	BYTE	opcode;		// opcode value to be modified by aProc()
	int		(*aProc)(int iOPS);	// address of routine to assemble this line, which gets passed the index into OPS[] for the opcode
	DWORD	flags;		// misc flags for various opcodes & addressing modes
} OPS_STRUC;

#define BLDOP(a,b,c,d) ((DWORD)((d)<<24)+(DWORD)((c)<<16)+(DWORD)((b)<<8)+(DWORD)(a))
OPS_STRUC OPS[]={
	{BLDOP('A','B','S',' '), 0302, &aAbs, 0},
	{BLDOP('A','D','B',' '), 0302, &aLDST, 0},
	{BLDOP('A','D','B','D'), 0322, &aLDST, 0},
	{BLDOP('A','D','M',' '), 0303, &aLDST, 0},
	{BLDOP('A','D','M','D'), 0323, &aLDST, 0},
	{BLDOP('A','I','F',' '),    0, &aAIF, 0},
	{BLDOP('A','N','M',' '), 0307, &aLDST, 0},
	{BLDOP('A','N','M','D'), 0327, &aLDST, 0},
	{BLDOP('A','R','P',' '),    0, &aArpDrp, 0000},
	{BLDOP('A','S','C',' '),    0, &aAscAsp, 0},	// flag==0 no parity on last char
	{BLDOP('A','S','P',' '),    0, &aAscAsp, 1},	// flag==1 parity on last char

	{BLDOP('B','C','D',' '), 0231, &aNoArgs, 0},
	{BLDOP('B','I','N',' '), 0230, &aNoArgs, 0},
	{BLDOP('B','S','Z',' '),    0, &aBsz, 0},
	{BLDOP('B','Y','T',' '),    0, &aValByt, 0},

	{BLDOP('C','L','B',' '), 0222, &aDR, 0},
	{BLDOP('C','L','E',' '), 0235, &aNoArgs, 0},
	{BLDOP('C','L','M',' '), 0223, &aDR, 0},
	{BLDOP('C','L','R',' '),    0, &aSetClr, 0},	// flag==0 for CLR label flag
	{BLDOP('C','M','B',' '), 0300, &aLDST, 0},
	{BLDOP('C','M','B','D'), 0320, &aLDST, 0},
	{BLDOP('C','M','M',' '), 0301, &aLDST, 0},
	{BLDOP('C','M','M','D'), 0321, &aLDST, 0},

	{BLDOP('D','C','B',' '), 0212, &aDR, 0},
	{BLDOP('D','C','E',' '), 0233, &aNoArgs, 0},
	{BLDOP('D','C','M',' '), 0213, &aDR, 0},
	{BLDOP('D','E','C',' '),    0, &aDEC, 0},
	{BLDOP('D','E','F',' '),    0, &aDef, 0},
	{BLDOP('D','R','P',' '), 0100, &aArpDrp, 0100},

	{BLDOP('E','I','F',' '),    0, &aEIF, 0},
	{BLDOP('E','L','B',' '), 0200, &aDR, 0},
	{BLDOP('E','L','M',' '), 0201, &aDR, 0},
	{BLDOP('E','L','I','F'),    0, &aELIF, 0},
	{BLDOP('E','L','S',' '),    0, &aELS, 0},
	{BLDOP('E','R','B',' '), 0202, &aDR, 0},
	{BLDOP('E','R','M',' '), 0203, &aDR, 0},

	{BLDOP('F','I','N',' '),    0, &aFin, 0},

	{BLDOP('G','T','O',' '),    0, &aGTO, 0},

	{BLDOP('H','E','D',' '),    0, &aHED, 0},

	{BLDOP('I','C','B',' '), 0210, &aDR, 0},
	{BLDOP('I','C','E',' '), 0234, &aNoArgs, 0},
	{BLDOP('I','C','M',' '), 0211, &aDR, 0},
	{BLDOP('I','N','C',' '), 0211, &aLNK, 1},	// just skips the filename, which gets processed in LoadSrc() flags=1 means INC

	{BLDOP('J','S','B',' '), 0316, &aJsb, 0},
	{BLDOP('J','C','Y',' '), 0373, &aJmp, 0},
	{BLDOP('J','E','N',' '), 0370, &aJmp, 0},
	{BLDOP('J','E','V',' '), 0363, &aJmp, 0},
	{BLDOP('J','E','Z',' '), 0371, &aJmp, 0},
	{BLDOP('J','L','N',' '), 0375, &aJmp, 0},
	{BLDOP('J','L','Z',' '), 0374, &aJmp, 0},
	{BLDOP('J','M','P',' '), 0360, &aJmp, 0},
	{BLDOP('J','N','C',' '), 0372, &aJmp, 0},
	{BLDOP('J','N','G',' '), 0364, &aJmp, 0},
	{BLDOP('J','N','O',' '), 0361, &aJmp, 0},
	{BLDOP('J','N','Z',' '), 0366, &aJmp, 0},
	{BLDOP('J','O','D',' '), 0362, &aJmp, 0},
	{BLDOP('J','P','S',' '), 0365, &aJmp, 0},
	{BLDOP('J','R','N',' '), 0377, &aJmp, 0},
	{BLDOP('J','R','Z',' '), 0376, &aJmp, 0},
	{BLDOP('J','Z','R',' '), 0367, &aJmp, 0},

	{BLDOP('L','D','B',' '), 0240, &aLDST, 010},	// opcode is for DR,AR - flag gets added to opcode for DR,=literal
	{BLDOP('L','D','B','D'), 0244, &aLDST, 014},	// opcode is for DR,AR - flag gets added to opcode for DR,=label - (flag+4) gets added to opcode for DR,XAR,label
	{BLDOP('L','D','B','I'), 0254, &aLDST, 014},	// opcode is for DR,AR - flag gets added to opcode for DR,=label - (flag+4) gets added to opcode for DR,XAR,label
	{BLDOP('L','D','M',' '), 0241, &aLDST, 010},	// opcode is for DR,AR - flag gets added to opcode for DR,=literal
	{BLDOP('L','D','M','D'), 0245, &aLDST, 014},	// opcode is for DR,AR - flag gets added to opcode for DR,=label - (flag+4) gets added to opcode for DR,XAR,label
	{BLDOP('L','D','M','I'), 0255, &aLDST, 014},	// opcode is for DR,AR - flag gets added to opcode for DR,=label - (flag+4) gets added to opcode for DR,XAR,label
	{BLDOP('L','L','B',' '), 0204, &aDR, 0},
	{BLDOP('L','L','M',' '), 0205, &aDR, 0},
	{BLDOP('L','N','K',' '),    0, &aLNK, 0},	// just skips the filename, which gets processed in LoadSrc(), flags=0 means LNK
	{BLDOP('L','O','C',' '),    0, &aLoc, 0},
	{BLDOP('L','R','B',' '), 0206, &aDR, 0},
	{BLDOP('L','R','M',' '), 0207, &aDR, 0},
	{BLDOP('L','S','T',' '),    0, &aLstUnl, 1},

	{BLDOP('N','A','M',' '),    0, &aNAM, 0},
	{BLDOP('N','C','B',' '), 0216, &aDR, 0},
	{BLDOP('N','C','M',' '), 0217, &aDR, 0},

	{BLDOP('O','C','T',' '),    0, &aValByt, 0},
	{BLDOP('O','R','B',' '), 0224, &aDRAR, 0},
	{BLDOP('O','R','G',' '), 0225, &aORG, 0},
	{BLDOP('O','R','M',' '), 0225, &aDRAR, 0},

	{BLDOP('P','A','D',' '), 0237, &aNoArgs, 1},
	{BLDOP('P','O','B','D'), 0340, &aPushPop, 0},	// +, - is 342
	{BLDOP('P','O','B','I'), 0350, &aPushPop, 0},	// +, - is 352
	{BLDOP('P','O','M','D'), 0341, &aPushPop, 0},	// +, - is 343
	{BLDOP('P','O','M','I'), 0351, &aPushPop, 0},	// +, - is 353
	{BLDOP('P','U','B','D'), 0344, &aPushPop, 0},	// +, - is 346
	{BLDOP('P','U','B','I'), 0354, &aPushPop, 0},	// +, - is 356
	{BLDOP('P','U','M','D'), 0345, &aPushPop, 0},	// +, - is 347
	{BLDOP('P','U','M','I'), 0355, &aPushPop, 0},	// +, - is 357

	{BLDOP('R','T','N',' '), 0236, &aNoArgs, 0},

	{BLDOP('S','A','D',' '), 0232, &aNoArgs, 0},
	{BLDOP('S','B','B',' '), 0304, &aLDST, 0},
	{BLDOP('S','B','B','D'), 0324, &aLDST, 0},
	{BLDOP('S','B','M',' '), 0305, &aLDST, 0},
	{BLDOP('S','B','M','D'), 0325, &aLDST, 0},
	{BLDOP('S','E','T',' '),    0, &aSetClr, 1},	// flag==1 for SET label flag
	{BLDOP('S','T','B',' '), 0242, &aLDST, 010},	// opcode is for DR,AR - flag gets added to opcode for DR,=literal
	{BLDOP('S','T','B','D'), 0246, &aLDST, 014},	// opcode is for DR,AR - flag gets added to opcode for DR,=label - (flag+4) gets added to opcode for DR,XAR,label
	{BLDOP('S','T','B','I'), 0256, &aLDST, 014},	// opcode is for DR,AR - flag gets added to opcode for DR,=label - (flag+4) gets added to opcode for DR,XAR,label
	{BLDOP('S','T','M',' '), 0243, &aLDST, 010},	// opcode is for DR,AR - flag gets added to opcode for DR,=literal
	{BLDOP('S','T','M','D'), 0247, &aLDST, 014},	// opcode is for DR,AR - flag gets added to opcode for DR,=label - (flag+4) gets added to opcode for DR,XAR,label
	{BLDOP('S','T','M','I'), 0257, &aLDST, 014},	// opcode is for DR,AR - flag gets added to opcode for DR,=label - (flag+4) gets added to opcode for DR,XAR,label

	{BLDOP('T','C','B',' '), 0214, &aDR, 0},
	{BLDOP('T','C','M',' '), 0215, &aDR, 0},
	{BLDOP('T','S','B',' '), 0220, &aDR, 0},
	{BLDOP('T','S','M',' '), 0221, &aDR, 0},

	{BLDOP('U','N','L',' '),    0, &aLstUnl, 0},

	{BLDOP('V','A','L',' '),    0, &aValByt, 0},

	{BLDOP('X','R','B',' '), 0226, &aDRAR, 0},
	{BLDOP('X','R','M',' '), 0227, &aDRAR, 0},

	{BLDOP('B','R','K',' '), 0346, &aNoArgs, 0},
};

void SetFormat(void) {

	OutForm.adrForm = HexOut ? pHexAdr : pOctAdr;
	OutForm.objForm = HexOut ? pHexObj : pOctObj;
	OutForm.objNum  = NumObj+1;
	OutForm.objW    = HexOut ? (5+OutForm.objNum*3) : (7+OutForm.objNum*4);
}

#define	SkipWhite(p)	while( *(p) && (*(p)==' ' || *(p)=='\t') ) ++(p)

///********************************
void UndefLabel(char *pLabel) {

	sprintf((char*)ScratBuf, "Undefined label: %s", pLabel);
	AsmError((char*)ScratBuf);
}

///********************************
LABEL *FindLocal(char *pLabel) {

	long	i;

	for(i=0; i<locCnt; i++) {
		if( !strcmp(pLabel, Locals[i].label) ) return &Locals[i];
	}
	return NULL;
}

///********************************
LABEL *FindGlobal(char *pLabel) {

	long	i;

	for(i=0; i<gloCnt; i++) {
		if( !strcmp(pLabel, Globals[i].label) ) {
			if( pass ) {
				if( Globals[i].type==DAD ) ++GDADcnt;
				else if( Globals[i].type==EQU ) ++GEQUcnt;
			}
			return &Globals[i];
		}
	}
	return NULL;
}

/***************************************
 * get a number (octal or dec or bcd) or return 0
 * NOTE: GetNum *MUST* leave pSRC pointing JUST past
 * any number-base-suffix (C, D, B), so that callers
 * (when defining DADs and EQUs) can "look back" and
 * save what format the number was specified in, so
 * it can be dumped in that format in the LABEL REFERENCE table.
 ******************************************/
BOOL	GetNum(DWORD *pDW) {

	char	*pc, *pch;
	DWORD	num;
	char	*saveSRC=pSRC;

	num = 0;
	SkipWhite(pSRC);
	if( (GetBNOT = (*pSRC=='~')) ) {	// if ones complement
		++pSRC;
		SkipWhite(pSRC);
	}
	if( *pSRC=='\'' ) {	// single-quote? if 'char'
		++pSRC;
		num = *pSRC++;
		if( *pSRC!='\'' ) AsmError("Invalid CHAR value");
		else ++pSRC;
	} else {
		if( !isdigit(*pSRC) ) {
			pSRC = saveSRC;	// restore in case of ~
			return FALSE;
		}
		pc = pSRC;
		for(; isdigit(*pSRC); pSRC++);
		pch = pSRC;
		for(; isxdigit(*pch); pch++);
		if( toupper(*pch)=='H' ) {	// hex
			while( isxdigit(*pc) ) {
				num = (num<<4) + (toupper(*pc) - '0');
				if( *pc>'9' ) num -= 7;
				++pc;
			}
			pSRC = ++pch;	// skip 'H'
		} else if( toupper(*pSRC)=='B' ) {	// binary
			while( *pc=='0' || *pc=='1' ) num = (num<<1) + (*pc++ - '0');
			++pSRC;
		} else if( toupper(*pSRC)=='D' ) {	// decimal
			num = atol(pc);
			++pSRC;
		} else if( toupper(*pSRC)=='C' ) {	// BCD
			while( *pc>='0' && *pc<='9' ) num = (num<<4) + (*pc++ - '0');
			++pSRC;
		} else {					// octal
			while( *pc>='0' && *pc<='7' ) num = (num<<3) + (*pc++ - '0');
			if( *pc=='8' || *pc=='9' ) AsmError("Invalid OCTAL value");
		}
	}
	if( GetBNOT ) num = ~num;
	*pDW = num;
	return TRUE;
}

///********************************
void MakeCond(char *label, DWORD value) {

	LABEL	*pLAB;

	pLAB = &Locals[locCnt++];
	pLAB->type = AIF;
	pLAB->refd = 0;
	strcpy(pLAB->label, label);
	pLAB->value = value;
	pLAB->unref = 0;
	pLAB->public = 0;
	pLAB->comment = NULL;
	++LAIFcnt;	// bump number of different AIF flags/labels encountered for dumping table at end
}

///********************************
void SetBPGM(BOOL isBP) {

	// isROM, isROM85, isROM87, isBPGM, isBPGM85, isBPGM87
	if( mach<0 ) {
		AsmError("SetBPGM before mach defined");
		return;
	}
	if( mach==MACH_85 ) {
		if( isROM<0 ) {	// if haven't yet set the 'is' AIFs
			isROM = !isBP;
			MakeCond("isROM", isROM);
			MakeCond("isROM85", isROM);
			MakeCond("isROM87", FALSE);
		}
		if( isBPGM<0 ) {
			isBPGM = isBP;
			MakeCond("isBPGM", isBPGM);
			MakeCond("isBPGM85", isBPGM);
			MakeCond("isBPGM87", FALSE);
		}
	} else if( mach==MACH_87 ) {
		if( isROM<0 ) {	// if haven't yet set the 'is' AIFs
			isROM = !isBP;
			MakeCond("isROM", isROM);
			MakeCond("isROM85", FALSE);
			MakeCond("isROM87", isROM);
		}
		if( isBPGM<0 ) {
			isBPGM = isBP;
			MakeCond("isBPGM", isBPGM);
			MakeCond("isBPGM85", FALSE);
			MakeCond("isBPGM87", isBPGM);
		}
	}
}

/********************************************************************************
 * Set MACH type and read GLOBALS files and convert GLOBAL entries to Globals[] *
 ********************************************************************************/
void SetMach(long m) {

	int		i, remlen, GlinesCnt, lnum;
	int    	fp;
	char	label[10], label2[10], nb, *pS, *pSaveSRC;
	char	**pGLines=NULL;
	BOOL	gn;
	LABEL	*pLAB;

	if( mach>=0 ) {
		if( mach!=m ) AsmError("Conflicting MODEL specification");
		return;
	}
	mach = m;
	if( mach==MACH_75 ) {
		HexOut ^= 1;	/// 75 defaults to hex
		if( NumObj==-1 ) NumObj = NumObjDef = 3;
	} else if( NumObj==-1 ) NumObj = NumObjDef = 3;
	SetFormat();

	// create HP85, HP87, or HP75 local conditional label
	MakeCond("HP85", mach==MACH_85);
	MakeCond("HP87", mach==MACH_87);
	MakeCond("HP75", mach==MACH_75);

	//======================================================================
	// LOAD and PROCESS GLOBALS FILE
	// open GLOBALS.TXT file (if it exists) for reading, allocate buffer, read the file in
	strcat(gloFileName, globfile[mach]);

	if( -1==(fp=open(gloFileName, _O_BINARY | _O_RDONLY)) ) {
//i = errno;
		printf("Unable to open GLOBALS file\r\n");
		exit(-1);
	}
	gloBufLen = _lseek(fp, 0, SEEK_END);
	// we allocate 1 more than the size of the file so that there's room for
	// a NUL byte at the end in case the idiot user has edited the file and
	// removed the eol seqquence from the end of the last line
	gloBuf = (char *)malloc(gloBufLen+1);
	if( gloBuf == NULL ) {
		printf("Error allocating GLOBALS buffer\r\n");
		exit(-1);
	}
	lseek(fp, 0, SEEK_SET);
	read(fp, (void*)gloBuf, gloBufLen);
	gloBuf[gloBufLen] = 0;
	close(fp);

	gloCnt = 0;

	// build array of line ptrs into GLOBALS file and replace \r and \n with NUL
	pGLO = gloBuf;
	GlinesCnt = 0;
	pGLines = (char **)malloc(65536*sizeof(char*));
	if( pGLines == NULL ) {
		printf("Error allocating pGLines buffer\r\n");
		exit(-1);
	}
	while( *pGLO ) {
		remlen = gloBufLen - (pGLO-gloBuf);
		// save ptr to start of line
		pGLines[GlinesCnt++] = pGLO;
		// skip to end of line
		while( remlen && *pGLO!='\r' && *pGLO!='\n' ) {
			++pGLO;
			--remlen;
		}
		// skip past end of line sequence and replace it with NUL chars
		if( remlen && *pGLO=='\r' ) {
			*pGLO++ = 0;
			--remlen;
		}
		if( remlen && *pGLO=='\n' ) {
			*pGLO++ = 0;
			--remlen;
		}
	}
	// now process the lines
	pSaveSRC = pSRC;	// save pSRC so we can set if for GetNum()
	for(lnum=0; lnum<GlinesCnt; lnum++) {
		pS = pGLines[lnum];
		SkipWhite(pS);
		if( *pS && *pS!='!' ) {	// lines starting with '!' are full-line comments, so ignore them
			// Get up to 8 chars of label
			for(i=0; i<8 && *pS && *pS!=' ' && *pS!='\t'; ) label[i++] = *pS++;
			label[i] = 0;
			if( i==8 && *pS!=' ' && *pS!='\t' ) printf("ERROR: GLOBAL label too long: %s", label);
			// skip WhiteSpace
			SkipWhite(pS);
			if( !strnicmp(pS, "EQU", 3) ) Globals[gloCnt].type = EQU;
			else if( !strnicmp(pS, "DAD", 3) ) Globals[gloCnt].type = DAD;
			else {
				printf("ERROR: GLOBAL unknown statement: %s", pGLines[lnum]);
				continue;
			}
			if( *pS ) ++pS;	// skip DAD or EQU
			if( *pS ) ++pS;	// skip DAD or EQU
			if( *pS ) ++pS;	// skip DAD or EQU
			SkipWhite(pS);
			pSRC = pS;
			gn = GetNum(&Globals[gloCnt].value);	// had to play tricks 'cause GetNum always uses pSRC. Could have changed GetNum, but this is the ONLY time we want to use it on something besides pSRC, so why pass another arg everytime? Sheesh. Who writes comments this long on ONE line???
			pS = pSRC;

			if( gn ) {
				Globals[gloCnt].refd = 0;	// hasn't been referenced yet
				nb = toupper(*(pS-1));	// get any number base specifier
				Globals[gloCnt].numbase = (nb=='H')?BASEH:((nb=='C')?BASEC:((nb=='D')?BASED:((nb=='B')?BASEB:BASEO)));
				strcpy(Globals[gloCnt].label, label);
				++gloCnt;
			} else {
				if( GetLabel(label2) && ((pLAB=FindLocal(label2))!=NULL || (pLAB=FindGlobal(label2))!=NULL) ) {	// local or global label
					++pLAB->refd;
					gn = BNOT(pLAB->value);
					Globals[gloCnt].refd = 0;	// hasn't been referenced yet
					Globals[gloCnt].numbase = pLAB->numbase;
					strcpy(Globals[gloCnt].label, label);
					++gloCnt;
				} else printf("ERROR: GLOBAL file label (%s) missing value%s", label, eolseq);
			}
		}
	}
	pSRC = pSaveSRC;	// restore pSRC
	free(gloBuf);
	gloBuf = NULL;
	gloBufLen = 0;
	free(pGLines);
}

/***************
 * Get a label *
 ***************/
// 'label' must be at least 10 bytes long
// if label found, SkipWhite() is done before return

/// WARNING! ASM85 doesn't allow a ',' in a label. CASM *MAY* have, so that may cause errors on original source code!
/// ASM85 doesn't allow a ',' so that VAL can have a list of labels.

BOOL GetLabel(char *label) {

	long	i;

	if( (GetBNOT = (*pSRC=='~')) ) {
		++pSRC;
		SkipWhite(pSRC);
	}
	if( (OneByteLabel = (*pSRC=='{')) ) {
		++pSRC;
		SkipWhite(pSRC);
	}
	if( *pSRC && *pSRC!='\r' && *pSRC!='\n' && *pSRC!=' ' && *pSRC!='\t' && *pSRC!=',' && *pSRC!='!' && (!OneByteLabel || *pSRC!='}') && !isdigit(*pSRC) ) {
		// Get up to 8 chars of label
		for(i=0; i<8 && *pSRC && (!OneByteLabel || *pSRC!='}') && *pSRC!=' ' && *pSRC!='\t' && *pSRC!=',' && *pSRC!='\r' && *pSRC!='\n'; ) label[i++] = *pSRC++;
		label[i] = 0;
		// skip WhiteSpace
		SkipWhite(pSRC);
		if( OneByteLabel ) {
			if( *pSRC=='}' ) ++pSRC;
			SkipWhite(pSRC);
		}
		return TRUE;
	}
	return FALSE;
}

///**************************************
void AsmError(char *err) {

	if( pass && !ErrBuf[0] ) {	// if not already a pending error
		sprintf(ErrBuf, "(%s line %d) ERROR: %s%s", curFileName, linenum-linenumBase+1, err, eolseq);
		++ErrCnt;
	}
}

///**************************************
void AsmWarning(char *err) {

	if( DoWarnings && pass && !WarnBuf[0] && !Okay ) {	// if not already a pending warning
		sprintf(ErrBuf, "(%s line %d) WARNING: %s%s", curFileName, linenum-linenumBase+1, err, eolseq);
		++WarnCnt;
	}
}

///**********************
void CheckEOLcomment() {

	SkipWhite(pSRC);
	if( *pSRC && *pSRC!='!' ) {
		AsmError("unexpected chars at EOL");
	}
}

///****************************
BOOL GetDRP() {

	DWORD	dw;

	SkipWhite(pSRC);
	if( toupper(*pSRC)=='R' ) ++pSRC;
	if( *pSRC=='^' ) {	// force drp to known value without a DRP instruction
		++pSRC;
		if( !GetNum(&dw) ) goto baddrp;
		if( drp>-1 && drp!=dw ) {
			AsmError("!xx DRP disagrees with known DRP");
			return FALSE;
		}
		drp = dw;
		return TRUE;
	}
	if( *pSRC=='#' ) {
		++pSRC;
		return TRUE;
	}
	if( *pSRC=='*' ) {
		++pSRC;
		dw = 1;
	} else if( !GetNum(&dw) ) {
baddrp:
		AsmError("Bad DRP format");
		return FALSE;
	}
	if( dw>077 ) {
		AsmError("Bad DRP");
		return FALSE;
	}
	if( drp!=dw ) {
		code[codelen++] = 0100+dw;
		drp = dw;
	}
	return TRUE;
}

///****************************
/// returns 0 if bad, 1 if R, 2 if X
long GetARP() {

	long	retval;
	DWORD	dw;

	retval = 1;
	SkipWhite(pSRC);
	if( toupper(*pSRC)=='R' ) {
		++pSRC;
	} else if( toupper(*pSRC)=='X' ) {
		++pSRC;
		retval = 2;
	}
	if( *pSRC=='^' ) {	// force arp to known value without an ARP instruction. NOTE: If used in an indexed instruction, it MUST be preceded by an 'X'
		++pSRC;
		if( !GetNum(&dw) ) goto badarp;
		if( arp>-1 && arp!=dw ) {
			AsmError("^xx ARP disagrees with known ARP");
			return FALSE;
		}
		arp = dw;
		return retval;
	}
	if( *pSRC=='#' ) {
		++pSRC;
		return retval;
	}
	if( *pSRC=='*' ) {
		++pSRC;
		dw = 1;
	} else if( !GetNum(&dw) ) {
badarp:
		AsmError("Bad ARP format");
		retval = 0;
	}
	if( dw>077 ) {
		AsmError("Bad ARP");
		retval = 0;
	}
	if( arp!=dw ) {
		code[codelen++] = dw;
		arp = dw;
	}
	return retval;
}

///************************
int aORG(int iOPS) {

	AsmError("ORG opcode not supported");
	return 1;
}

///************************
int aSetClr(int iOPS) {

	char	label[10];
	LABEL	*pLAB;

	SkipWhite(pSRC);
	if( GetLabel(label) ) {
		if( (pLAB=FindLocal(label))==NULL ) {	// not defined yet
			pLAB = &Locals[locCnt++];
			pLAB->type = AIF;
			pLAB->refd = 0;
			strcpy(pLAB->label, label);
			++LAIFcnt;	// bump number of different AIF flags/labels encountered for dumping table at end
		} else if( pLAB->type!=AIF ) {
			AsmError("Label not SET/CLR flag");
			return 1;
		} else {
			++LAIFcnt;
		}
		pLAB->value = OPS[iOPS].flags;
		pLAB->unref = UnRef;
		pLAB->public = 0;
		pLAB->comment = NULL;
	}
	return 0;
}

///************************
int aAIF(int iOPS) {

	char	label[10];
	LABEL	*pLAB;

	SkipWhite(pSRC);
	if( AIFcnt>=255 ) {
		AsmError("Too many nested AIFs");

		return 1;
	}
	if( *pSRC!='!' ) {
		if( GetLabel(label) ) {
			if( (pLAB=FindLocal(label))==NULL ) {
				AsmError("AIF/SET/CLR label not found");
				return 1;
			}
			AIFstack[AIFcnt++] = isAIF | (isAIFdone<<8);
			if( pass ) ++pLAB->refd;
			if( isAIF ) {
				if( !BNOT(pLAB->value) ) isAIF = isAIFdone = 0;	// if we were assembling when we hit the AIF, but the AIF value is false, we haven't found the right case (ELS/ELSIF) yet
				else isAIFdone = 1;	// if we were assembling, and the AIF value is TRUE, then we're doing the immediate AIF block, so we're done for any ELS/ELSIF blocks until EIF
			} else isAIFdone = 1;	// if we weren't assembling when we hit the AIF, we're immediately done with EVERYTHING within this AIF
			AIFstmt = 1;
		}
	}
	return 0;
}

///************************
int aELIF(int iOPS) {

	char	label[10];
	LABEL	*pLAB;

	SkipWhite(pSRC);
	if( 0==AIFcnt ) {
		AsmError("ELS without matching AIF");
		return 1;
	}
	if( *pSRC!='!' ) {
		if( GetLabel(label) ) {
			if( (pLAB=FindLocal(label))==NULL ) {
				AsmError("AIF/SET/CLR label not found");
				return 1;
			}
//			AIFstack[AIFcnt++] = isAIF | (isAIFdone<<8);
			if( pass ) ++pLAB->refd;

			if( !isAIFdone ) {
				if( !BNOT(pLAB->value) ) isAIF = 0;
				else isAIFdone = isAIF = 1;
			} else isAIF = 0;
			AIFstmt = 1;
		}
	}
	return 0;
}

///************************
int aELS(int iOPS) {

	if( 0==AIFcnt ) {
		AsmError("ELS without matching AIF");

		return 1;
	}
	if( !isAIFdone ) {
		isAIF = isAIFdone = 1;
	} else isAIF = 0;
	AIFstmt = 1;
	return 0;
}

///************************
int aEIF(int iOPS) {

	if( 0==AIFcnt ) {
		AsmError("EIF without matching AIF");

		return 1;
	}
	isAIF = AIFstack[--AIFcnt] & 0xFF;
	isAIFdone = (AIFstack[AIFcnt]>>8);
	AIFstmt = 1;
	return 0;
}

///************************
int aValByt(int iOPS) {

	char	label[10];
	DWORD	val=0;
	LABEL	*pLAB;


	do {
		SkipWhite(pSRC);
		if( GetNum(&val) ) {
			if( (!GetBNOT && val>255) || (GetBNOT && BNOT(val)>255) ) {
				AsmError("Number >255");
				return 1;
			}
		} else if( GetLabel(label) ) {
			if( (pLAB=FindLocal(label))!=NULL || (pLAB=FindGlobal(label))!=NULL ) {	// local or global label
				if( pass ) {
					++pLAB->refd;
					val = BNOT(pLAB->value);
					if( pLAB->type!=EQU || (pLAB->value & 0x0FF00) ) {
						AsmError("Expected EQU label with value <256");
						return 1;
					}
				}
			} else if( pass ) {
				UndefLabel(label);
				return 1;
			}
		} else {
			AsmError("Number or EQU label expected");
			return 1;
		}
		code[codelen++] = val & 0x0FF;
		SkipWhite(pSRC);
		if( *pSRC!=',' ) break;
		++pSRC;
	} while( TRUE );
	return 0;
}

///************************
int aBsz(int iOPS) {

	DWORD	dw;
	int		zcnt, i;

	SkipWhite(pSRC);
	if( !GetNum(&dw) ) {
		AsmError("Number expected");
		return 1;
	}
	if( dw>8192 ) {
		AsmError("BSZ > 8191");
		return 1;
	}
	zcnt = min(3, dw);
	BSZcnt = dw-zcnt;
	for(i=0; i<zcnt; i++) code[codelen++] = 0;

	return 0;
}

///************************
int aDef(int iOPS) {

	DWORD	val;
	char	label[10];
	LABEL	*pLAB;

	SkipWhite(pSRC);
	if( !GetLabel(label) ) {
		AsmError("Label expected");
		return 1;
	}
	if( (pLAB=FindLocal(label))!=NULL || (pLAB=FindGlobal(label))!=NULL ) {	// local or global label
		++pLAB->refd;
		val = BNOT(pLAB->value);
	} else {	// undefined label. If pass==0, treat as forward reference local label, else throw undefined label error
		val = 0;
		if( pass!=0 ) {
			code[codelen++] = 0;
			code[codelen++] = 0;
			UndefLabel(label);
			return 1;
		}
	}
	code[codelen++] = val & 0x0FF;
	code[codelen++] = ((val>>8) & 0x0FF);
	return 0;
}

///************************
/// Chew up decimal numbers and bytes with the corresponding BCD digits
int aDEC(int iOPS) {

	BYTE	digs[16], d;
	int		di;

	do {
		SkipWhite(pSRC);
		if( !isdigit(*pSRC) ) {
			AsmError("Number expected");
			return 1;
		}
		for(di=0; isdigit(*pSRC); ++di, ++pSRC) digs[di] = *pSRC-'0';
		while( di>0 ) {
			d = digs[--di];
			if( di ) d |= (digs[--di]<<4);
			code[codelen++] = d;
		}
		SkipWhite(pSRC);
		if( *pSRC!=',' ) break;
		++pSRC;
	} while( TRUE );
	return 0;
}

///************************
int aAscAsp(int iOPS) {

	BYTE	last=0, lastch;
	DWORD	len;

	if( OPS[iOPS].flags ) last = 0x80;
	SkipWhite(pSRC);
	if( isdigit(*pSRC) ) {		// #,unquoted string
		GetNum(&len);
		SkipWhite(pSRC);
		if( *pSRC!=',' ) {
			AsmError("Comma expected");
			return 1;
		}
		++pSRC;
		if( len ) {
			while( --len && *pSRC ) code[codelen++] = *pSRC++;
			if( *pSRC ) code[codelen++] = last | *pSRC++;
		}
		while( *pSRC ) ++pSRC;	// skip rest of line, we've got the chars asked for and CASM ignored the rest
	} else if( *pSRC=='"' )	{	// quoted string
		++pSRC;
		if( *pSRC && *pSRC!='"') {
			lastch = *pSRC++;
			while( *pSRC && *pSRC!='"' ) {
				code[codelen++] = lastch;
				lastch = *pSRC++;
			}
			code[codelen++] = last | lastch;
		} else if( *pSRC==0 ) {
			AsmError("Unexpected end of line");
			return 1;
		}
		if( *pSRC=='"' ) ++pSRC;
	}
	return 0;
}

///************************
int aLoc(int iOPS) {

	DWORD	val;
	char	label[10];
	LABEL	*pLAB;

	if( baseAddr<0 ) {
		AsmError("LOC in non-ABS code");
		return 1;
	}
	SkipWhite(pSRC);
	if( !GetNum(&val) ) {
		if( !GetLabel(label) ) {
			AsmError("Address or label expected");
			return 1;
		}
		if( (pLAB=FindLocal(label))!=NULL || (pLAB=FindGlobal(label))!=NULL ) {	// local or global label
			++pLAB->refd;
			val = pLAB->value;
		} else {	// throw undefined label error
			AsmError("Label must be pre-defined");
			return 1;
		}
	}
	if( val < nextAddr ) {
		AsmError("Can't LOC backward");
		return 1;
	}
	if( val > nextAddr ) {
		BSZcnt = val-nextAddr;
	}
	return 0;
}

///************************
int aNoArgs(int iOPS) {

	BYTE	oc=OPS[iOPS].opcode;

	if( OPS[iOPS].flags ) {	// it's PAD, invalidate ARP/DRP
		drp = arp = dcm = -1;
	}
	if( oc==0231 ) dcm = 1;	// BCD
	if( oc==0230 ) dcm = 0;	// BIN (for relative BPGM GTO check)
	if( oc==0336 ) {	// BRK, spit WARNING
		AsmWarning("BRK only works in the Emulator, it's a NO-OP on real hardware!");
	}
	SkipWhite(pSRC);
	code[codelen++] = oc;
	return 0;
}

///************************
int aArpDrp(int iOPS) {

	if( OPS[iOPS].flags ) {
		drp = -1;
		GetDRP();
	} else {
		arp = -1;
		GetARP();
	}
	return 0;
}

///************************
int aLstUnl(int iOPS) {

	ListOn = OPS[iOPS].flags;

	return 0;
}

///************************
int aJsb(int iOPS) {

	DWORD	val;
	char	label[10];
	LABEL	*pLAB;

	SkipWhite(pSRC);
	if( *pSRC=='=' ) {
		++pSRC;
		code[codelen++] = OPS[iOPS].opcode;
		goto jsbcom;
	} else if( GetARP() ) {
		SkipWhite(pSRC);
		if( *pSRC!=',' ) goto synerr;
		++pSRC;
		if( *pSRC=='=' ) ++pSRC;	// skip optional '=' on the indexed JSB
		code[codelen++] = OPS[iOPS].opcode & ~010;
jsbcom:
		drp = arp = dcm = -1;
		SkipWhite(pSRC);
		if( !GetLabel(label) ) {
			AsmError("Missing label reference");
			return 1;
		}
		if( (pLAB=FindLocal(label))!=NULL || (pLAB=FindGlobal(label))!=NULL ) {	// local or global
			++pLAB->refd;
			val = pLAB->value;
		} else {	// undefined label. If pass==0, treat as forward reference local label, else throw undefined label error
			val = 0;
			if( pass!=0 ) {
				code[codelen++] = 0;
				code[codelen++] = 0;
				UndefLabel(label);
				return 1;
			}
		}
		code[codelen++] = val & 0x0FF;
		code[codelen++] = (val >> 8) & 0x0FF;
	} else {
synerr:
		AsmError("Syntax");
		return 1;
	}
	return 0;
}

///************************
int aJmp(int iOPS) {

	DWORD	val;
	char	label[10];
	LABEL	*pLAB;
	long	delta, roff=0;
	BYTE	oc=OPS[iOPS].opcode;
	int		retval=0, ATdir, ATcnt, ATbyte, ATbit;

	SkipWhite(pSRC);
	if( *pSRC=='@' ) {
		++pSRC;
		if( *pSRC=='+' || toupper(*pSRC)=='F' ) ATdir = 1;
		else if( *pSRC=='-' || toupper(*pSRC)=='B' ) ATdir = -1;
		else {
			AsmError("Invalid @ label");
			++retval;
		}
		if( !retval ) {
			// count & skip the +/F's or -/B's
			for(ATcnt=0; (ATdir==1 && (*pSRC=='+' || toupper(*pSRC)=='F')) || (ATdir==-1 && (*pSRC=='-' || toupper(*pSRC)=='B')); ATcnt++, pSRC++) ;
			if( pass ) {
				if( ATdir==1 ) {	/// going forward
					ATbyte = (nextAddr)/8;
					ATbit  = ((nextAddr) & 7);
					while( ATcnt-- ) {
						while( ++ATbit<8 ) if( ATlabels[ATbyte] & (1<<ATbit) ) break;	// finish out searching first byte
						if( ATbit >= 8 ) {
							while( ++ATbyte<8192 ) if( ATlabels[ATbyte] ) break;		// search for next byte with a bit set
							if( ATbyte >= 8192 ) {
								AsmError("Missing @ target");
								++retval;
								goto pushfin;
							}
							for(ATbit=0; !(ATlabels[ATbyte] & (1<<ATbit)); ATbit++) ;	// find first bit set
						}
					}
					val = 8*ATbyte+ATbit;
				} else {			/// going backward
					ATbyte = (nextAddr+1)/8;
					ATbit  = ((nextAddr+1) & 7);
					while( ATcnt-- ) {
						while( --ATbit>=0 ) if( ATlabels[ATbyte] & (1<<ATbit) ) break;	// finish out searching first byte
						if( ATbit < 0 ) {
							while( --ATbyte>=0 ) if( ATlabels[ATbyte] ) break;		// search for next byte with a bit set
							if( ATbyte < 0 ) {
								AsmError("Missing @ target");
								++retval;
								goto pushfin;
							}
							for(ATbit=7; !(ATlabels[ATbyte] & (1<<ATbit)); ATbit--) ;	// find first bit set
						}
					}
					val = 8*ATbyte+ATbit;
				}
				goto gotval;
			}
		}
	} else if( !GetLabel(label) ) {
		AsmError("Label expected");
		++retval;
	} else if( pass ) {
//if( nextAddr==074077 || nextAddr==074037 ) {
//	val = 0;
//}
		if( (pLAB=FindLocal(label))==NULL ) {	// local only

			UndefLabel(label);
			++retval;
		} else {
			++pLAB->refd;
			val = pLAB->value;
gotval:
			if( val < nextAddr ) {	// jump backward
				delta = ((long)nextAddr+1-(long)val+1);
				if( delta > 128 ) {
					sprintf((char*)ScratBuf,"Jump out of range (by %dD bytes)", (int)(delta-128));
					AsmError((char*)ScratBuf);
					++retval;
				} else roff = (-delta);
			} else {	// jump forward
				delta = ((long)val-(long)nextAddr-2);
				if( delta > 127 ) {
					sprintf((char*)ScratBuf,"Jump out of range (by %dD bytes)", (int)(delta-127));
					AsmError((char*)ScratBuf);
					++retval;
				} else roff = delta;
			}
		}
	} else roff = 0;
pushfin:
	code[codelen++] = oc;
	code[codelen++] = (BYTE)(roff & 0x0FF);
	return retval;
}

///************************
int aPushPop(int iOPS) {

	BYTE	oc=OPS[iOPS].opcode;

	if( !GetDRP() ) {
		// error already output, do nothing else
		return 1;
	}
	SkipWhite(pSRC);
	if( *pSRC!=',' ) {
		AsmError("',' expected");

		return 1;
	}
	++pSRC;
	SkipWhite(pSRC);
	// next should be + or -
	if( toupper(*pSRC)=='-' ) {
		oc += 2;	// turn + opcode into -
	} else if( *pSRC!='+' ) {
		AsmError("Unknown addressing mode");
		return 1;
	}
	++pSRC;	// skip '+' or '-'
	if( !GetARP() ) return 1;
	code[codelen++] = oc;
	return 0;
}

///************************
int aDR(int iOPS) {

	BYTE	oc=OPS[iOPS].opcode;

	if( !GetDRP() ) {
		// error already output, do nothing else
		return 1;
	}
	code[codelen++] = oc;
	return 0;
}

///************************
int aLDST(int iOPS) {

	BYTE	oc=OPS[iOPS].opcode;
	char	label[10];
	DWORD	val;
	long	mbcnt;
	LABEL	*pLAB;
	BOOL	imm=(OPS[iOPS].opname>>24)==' ';
	BOOL	ldst=(BOOL)(OPS[iOPS].flags);	// 0 for Cm,Ad,Su,An. (014 or 010) for Ld,St.

	if( !GetDRP() ) {
		// error already output, do nothing else
		return 1;
	}

	if( oc & 1 ) {	// if multi-byte operation, set mbcnt for ,=literal operations
		if( drp<0 || (drp<040 && !(drp & 1)) ) mbcnt = 2;
		else if( drp>=040 ) mbcnt = 8 - (drp&7);
		else mbcnt = 1;
	} else mbcnt = 1;

	SkipWhite(pSRC);
	if( *pSRC!=',' ) {
synerr:;
		AsmError("Syntax");
		return 1;
	}
	++pSRC;	// skip comma
	SkipWhite(pSRC);




	if( *pSRC=='=' ) {	// DR,=literal or DR,=label
		++pSRC;	// skip '='
		SkipWhite(pSRC);
		if( imm ) {	// LD/ST DR,=literal values, could be #s, EQU, or DAD
			if( !ldst ) oc |= 0010;	// modify Cm,Ad,Su,An opcodes
			else oc += OPS[iOPS].flags;	// modify opcode for DR,=literal and DR,=label
			code[codelen++] = oc;
			do {
				if( !GetNum(&val) ) {	// not number, try label
					if( !GetLabel(label) ) {
						if( oc==0251 || oc==0253 || oc==0311 ) break;	// if LDM/CMM R#,= with no arguments (arguments may be on following lines)
						goto synerr;
					}
		// if find label, decide whether to output 1 or 2 bytes depending upon mbcnt
					if( (pLAB=FindLocal(label))!=NULL || (pLAB=FindGlobal(label))!=NULL ) {	// local or global
						++pLAB->refd;
						val = BNOT(pLAB->value);
						if( drp==4 && (oc & 0xE2)==0xA0 ) --val;	// if LD R4,= instruction, need one less as it's going to get incremented by the CPU as the load completes
						code[codelen++] = val & 0x0FF;
						--mbcnt;
						if( !OneByteLabel && mbcnt ) {
							code[codelen++] = (val >> 8) & 0x0FF;
							--mbcnt;
						}
					} else {	// undefined label. If pass==0, treat as forward reference local label, else throw undefined label error
						val = 0;
						if( pass!=0 ) {
							code[codelen++] = 0;
							code[codelen++] = 0;
							UndefLabel(label);
							return 1;
						}
						code[codelen++] = val & 0x0FF;
						--mbcnt;
						if( !OneByteLabel && mbcnt ) {
							code[codelen++] = (val >> 8) & 0x0FF;
							--mbcnt;
						} else if( drp>=0 ) {
toomanyargs:;
							AsmError("Arg list too long for register boundary");
							return 1;
						}
					}
				} else {
					// val = number
					// decide whether to output 1 or 2 bytes based on val>255 and B or M and drp
					if( drp==4 && (oc & 0xE2)==0xA0 ) --val;	// if LD R4,= instruction, need one less as it's going to get incremented by the CPU as the load completes
					code[codelen++] = val & 0x0FF;
					--mbcnt;
					if( mbcnt && ((!GetBNOT && val>255) || (GetBNOT && 0x0FF00!=(val & 0x0FF00))) ) {	// if there's registers to go and number was >255
						code[codelen++] = (val>>8) & 0x0FF;
						--mbcnt;
					}
				}
				SkipWhite(pSRC);
				if( *pSRC!=',' ) break;
				++pSRC;
				if( mbcnt <= 0 && drp!=1 && drp!=-1 ) goto toomanyargs;
			} while( TRUE );
			if( mbcnt && drp!=1 && drp!=-1 ) {
				AsmWarning("Not enough arguments");
			}
		} else {	// LD/ST DIRECT or INDIRECT, can only be a LABEL (local or global)
			if( ldst ) oc += OPS[iOPS].flags;	// modify opcode for DR,=literal and DR,=label
			goto finlabel;
		}
	} else {
		if( !GetARP() ) return 1;
		if( !imm && !ldst ) oc |= 0010;	// Cm,Ad,Su,An opcode DIRECT, so add 010 to opcode for R,R
		SkipWhite(pSRC);
		if( *pSRC==',' ) {	// must be indexed
			if( ldst ) {	// indexed is only allowed for LD and ST
				++pSRC;		// skip the comma
				oc += (OPS[iOPS].flags+4);	// modify opcode for DR,XAR,label
				SkipWhite(pSRC);
finlabel:
				code[codelen++] = oc;
				if( !GetLabel(label) ) goto synerr;
				if( (pLAB=FindLocal(label))!=NULL || (pLAB=FindGlobal(label))!=NULL ) {	// local or global
					++pLAB->refd;
					val = pLAB->value;
				} else {	// undefined label. If pass==0, treat as forward reference local label, else throw undefined label error
					val = 0;
					if( pass!=0 ) {
						code[codelen++] = 0;
						code[codelen++] = 0;
						UndefLabel(label);
						return 1;
					}
				}
				code[codelen++] = val & 0x0FF;
				code[codelen++] = (val >> 8) & 0x0FF;
			} else goto synerr;
		} else code[codelen++] = oc;
	}
#if 0
	if( toupper(*pSRC)=='R' || *pSRC=='^' ) {	// DR,AR
		if( !imm && !ldst ) oc |= 0010;	// Cm,Ad,Su,An opcode DIRECT, so add 010 to opcode for R,R
        if( !GetARP() ) return 1;	// error message already spit in GetARP()
		code[codelen++] = oc;
	} else if( *pSRC=='=' ) {	// DR,=literal or DR,=label
		++pSRC;	// skip '='
		SkipWhite(pSRC);
		if( imm ) {	// LD/ST DR,=literal values, could be #s, EQU, or DAD
			if( !ldst ) oc |= 0010;	// modify Cm,Ad,Su,An opcodes
			else oc += OPS[iOPS].flags;	// modify opcode for DR,=literal and DR,=label
			code[codelen++] = oc;
			do {
				if( !GetNum(&val) ) {	// not number, try label
					if( !GetLabel(label) ) {
						if( oc==0251 || oc==0253 || oc==0311 ) break;	// if LDM/CMM R#,= with no arguments (arguments may be on following lines)
						goto synerr;
					}
		// if find label, decide whether to output 1 or 2 bytes depending upon mbcnt
					if( (pLAB=FindLocal(label))!=NULL || (pLAB=FindGlobal(label))!=NULL ) {	// local or global
						++pLAB->refd;
						val = BNOT(pLAB->value);
						if( drp==4 && (oc & 0xE2)==0xA0 ) --val;	// if LD R4,= instruction, need one less as it's going to get incremented by the CPU as the load completes
						code[codelen++] = val & 0x0FF;
						--mbcnt;
						if( !OneByteLabel && mbcnt ) {
							code[codelen++] = (val >> 8) & 0x0FF;
							--mbcnt;
						}
					} else {	// undefined label. If pass==0, treat as forward reference local label, else throw undefined label error
						val = 0;
						if( pass!=0 ) {
							code[codelen++] = 0;
							code[codelen++] = 0;
							UndefLabel(label);
							return 1;
						}
						code[codelen++] = val & 0x0FF;
						--mbcnt;
						if( !OneByteLabel && mbcnt ) {
							code[codelen++] = (val >> 8) & 0x0FF;
							--mbcnt;
						} else if( drp>=0 ) {
toomanyargs:;
							AsmError("Arg list too long for register boundary");
							return 1;
						}
					}
				} else {
					// val = number
					// decide whether to output 1 or 2 bytes based on val>255 and B or M and drp
					if( drp==4 && (oc & 0xE2)==0xA0 ) --val;	// if LD R4,= instruction, need one less as it's going to get incremented by the CPU as the load completes
					code[codelen++] = val & 0x0FF;
					--mbcnt;
					if( mbcnt && ((!GetBNOT && val>255) || (GetBNOT && 0x0FF00!=(val & 0x0FF00))) ) {	// if there's registers to go and number was >255
						code[codelen++] = (val>>8) & 0x0FF;
						--mbcnt;
					}
				}
				SkipWhite(pSRC);
				if( *pSRC!=',' ) break;
				++pSRC;
				if( mbcnt <= 0 && drp!=1 && drp!=-1 ) goto toomanyargs;
			} while( TRUE );
			if( mbcnt && drp!=1 && drp!=-1 ) {
				AsmWarning("Not enough arguments");
			}
		} else {	// LD/ST DIRECT or INDIRECT, can only be a LABEL (local or global)
			if( ldst ) oc += OPS[iOPS].flags;	// modify opcode for DR,=literal and DR,=label
			code[codelen++] = oc;
			goto finlabel;
		}
	} else if( ldst && toupper(*pSRC)=='X' ) {	// DR,XAR,label
		oc += (OPS[iOPS].flags+4);	// modify opcode for DR,XAR,label
		if( !GetARP() ) goto synerr;
		code[codelen++] = oc;
		SkipWhite(pSRC);
		if( *pSRC!=',' ) goto synerr;
		++pSRC;	// skip comma
		SkipWhite(pSRC);
finlabel:
		if( !GetLabel(label) ) goto synerr;
		if( (pLAB=FindLocal(label))!=NULL || (pLAB=FindGlobal(label))!=NULL ) {	// local or global
			++pLAB->refd;
			val = pLAB->value;
		} else {	// undefined label. If pass==0, treat as forward reference local label, else throw undefined label error
			val = 0;
			if( pass!=0 ) {
				code[codelen++] = 0;
				code[codelen++] = 0;
				UndefLabel(label);
				return 1;
			}
		}
		code[codelen++] = val & 0x0FF;
		code[codelen++] = (val >> 8) & 0x0FF;
	} else goto synerr;
#endif
	return 0;
}

///************************
int aDRAR(int iOPS) {

	BYTE	oc=OPS[iOPS].opcode;

	if( !GetDRP() ) return 1;
	SkipWhite(pSRC);
	if( *pSRC!=',' ) {
		AsmError("Syntax");
		return 1;
	}
	++pSRC;	// skip comma
    GetARP();
    code[codelen++] = oc;
	return 0;
}

///************************
int aGTO(int iOPS) {

	DWORD	val, opcoff;
	char	label[10];
	LABEL	*pLAB;

	SkipWhite(pSRC);
	if( drp!=4 ) code[codelen++] = 0104;	// set DRP for LDM R4,=label-1
	opcoff = codelen;
	code[codelen++] = 0251;	// LDM R,=literal
	if( !GetLabel(label) ) {
		AsmError("Missing label reference");
		return 1;
	}
	if( (pLAB=FindLocal(label))!=NULL || (pLAB=FindGlobal(label))!=NULL ) {	// local or global
		++pLAB->refd;
		val = pLAB->value;
	} else {	// undefined label. If pass==0, treat as forward reference local label, else throw undefined label error
		val = 1;
		if( pass!=0 ) {
			code[codelen++] = 0;
			code[codelen++] = 0;
			UndefLabel(label);
			return 1;
		}
	}
	if( seenNAM && baseAddr==-1 && pass!=0 && pLAB->type==LOC ) {	// if relative BPGM on second pass
		if( dcm ) AsmWarning("GTO uses ADM, state of BIN unknown, may need to a BIN instruction");
		val -= (nextAddr+codelen+1);
		code[opcoff] = 0313;	// (change opcode from LDM to ADM)
	}
	--val;	// because CPU will inc R4 after load instruction
	code[codelen++] = val & 0x0FF;
	code[codelen++] = (val >> 8) & 0x0FF;
	return 0;
}

/************************************************
 * Loads the initial source file, then processes
 * the file(s), turning EOL-chars into NULs, and
 * scanning for INC and LNK opcodes, and merging
 * those files in, so that the memory version of
 * the source code is one continuous block.
 ************************************************/
void LoadSrc(char *fname) {

	int		fp, expBL, i, j, isLNK;
	char	*expB;

	linesCnt = 0;

	if( srcBuf!=NULL ) free(srcBuf);
	if( -1==(fp=open(fname, _O_BINARY | _O_RDONLY)) ) {
		printf("Error reading SOURCE FILE %s\r\n", fname);
		exit(-1);
	}
	srcBufLen = lseek(fp, 0, SEEK_END);
	srcBuf = (char *)malloc(srcBufLen+1);
	if( srcBuf == NULL ) {
		printf("Error allocating SOURCE buffer\r\n");
		exit(-1);
	}
	lseek(fp, 0, SEEK_SET);
	read(fp, (void*)srcBuf, srcBufLen);
	srcBuf[srcBufLen] = 0;
	close(fp);

	strcpy(curFileName, fname);

	// set ptrs to lines, NUL-term the line, scan for INC/LNK and include those files
	i = 0;
	while( srcBuf[i] ) {
		pLines[linesCnt++] = srcBuf+i;

		/// now check for INC or LNK opcodes and handle them if present
		if( srcBuf[i] && srcBuf[i]!='!' && srcBuf[i]!='\r' && srcBuf[i]!='\n' ) {	// if MAY be actual code
			if( srcBuf[i]!=' ' && srcBuf[i]!='\t' ) {	// not whitespace, must be label
				while( srcBuf[i] && srcBuf[i]!=' ' && srcBuf[i]!='\t' && srcBuf[i]!='\r' && srcBuf[i]!='\n' ) ++i;	// skip label
			}
			while( srcBuf[i]==' ' || srcBuf[i]=='\t' ) ++i;	// skip whitespace after label
			if( (!(isLNK=0) && !strnicmp(srcBuf+i, "INC", 3)) || ((isLNK=1) && !strnicmp(srcBuf+i, "LNK", 3)) ) {	// if blending in another file, set isLNK to 0 for INC or 1 for LNK, then...
				i += 3;	// skip INC/LNK opcode
				while( srcBuf[i]==' ' || srcBuf[i]=='\t' ) ++i;	// skip whitespace after opcode
				if( srcBuf[i]!='"' ) {
					for(j=i; srcBuf[j] && srcBuf[j]!='\r' && srcBuf[j]!='\n'; j++) ;	// skip to EOL
					srcBuf[j] = 0;	// NUL-term for next line
					printf("Quoted filename expected on INC/LNK opcode: %s", srcBuf+i);
					exit(-1);
				}
				++i;	// skip quote
				for(j=0; srcBuf[i] && srcBuf[i]!='"'; ) inclnkFileName[j++] = srcBuf[i++];	// copy filename
				inclnkFileName[j] = 0;	// NUL-term it
				while( srcBuf[i] && srcBuf[i]!='\r' && srcBuf[i]!='\n' ) i++;	// skip rest of line, keeping it for assembly and listing
				srcBuf[i++] = 0;	// NUL-term the INC/LNK line
				if( isLNK ) srcBufLen = i;	// it's LNK, so blow away the rest of the current file
				// open the INC/LNK file and get it's size
				if( -1==(fp=open(inclnkFileName, _O_BINARY | _O_RDONLY)) ) {
					printf("Error reading SOURCE FILE %s\r\n", inclnkFileName);
					exit(-1);
				}
				expBL = lseek(fp, 0, SEEK_END) + (isLNK?0:strlen(EINC));	// if INC, add for " EINC\r\n" pseudo-op we'll add for popping the filename/line number from incSTK (for error reporting)
				expB = (char *)malloc(expBL+srcBufLen+1);	// allocate a new buffer big enough for the previous src and the INC/LNK source (plus NUL)
				if( expB == NULL ) {
					printf("Error allocating SOURCE buffer\r\n");
					exit(-1);
				}
				lseek(fp, 0, SEEK_SET);		// seek back to the start of the INC/LNK file
				memcpy(expB, srcBuf, i);	// copy the pre-INC/LNK part of the previous src into the new buffer
				read(fp, (void*)(expB+i), expBL);	// read the INC/LNK src into the buffer (may request 7 more than there is, if INC, but who cares?)
				memcpy(expB+i+expBL, srcBuf+i, srcBufLen-i);	// copy the post-INC/LNK part of the previous src (if any) into the new buffer
				expB[expBL+srcBufLen] = 0;	// NUL-term the new buffer
				close(fp);	// close the INC/LNK file

				if( !isLNK ) memcpy(expB+i+expBL-strlen(EINC), EINC, strlen(EINC));	// add the "end-of-inc, pop the incSTK" pseudo-op

				for(j=0; j<linesCnt; j++) pLines[j] = expB + (pLines[j]-srcBuf);	// update all of the pLines[] ptrs to the new buf

				free(srcBuf);		// free the old buf
				srcBuf = expB;		// set srcBuf to the new buffer
				srcBufLen += expBL;	// adjust srcBufLen

				continue;
			}
		}
		/// else handle non-INC, non-LNK lines
		while( srcBuf[i] && srcBuf[i]!='\n' && srcBuf[i]!='\r' ) ++i;
		// now NUL-term ONE line (which may be "\r", "\n", "\r\n", so only NUL one of each)
		if( srcBuf[i]=='\r' ) srcBuf[i++] = 0;
		if( srcBuf[i]=='\n' ) srcBuf[i++] = 0;
	}
	pSRC = srcBuf;
}

///*******************
int aLNK(int iOPS) {

	long	j;

	SkipWhite(pSRC);
	++pSRC;	// skip " (which has to be there, or we'd have failed in LoadSrc and exited
	for(j=0; *pSRC && *pSRC!='"'; ) inclnkFileName[j++] = *pSRC++;	// copy filename
	inclnkFileName[j] = 0;	// NUL-term it
	++pSRC;	// skip closing "

	if( OPS[iOPS].flags ) {	// INC
		if( incSTKcnt>9 ) AsmError("Too many nested INCs");
		else {
			strcpy(incSTKname[incSTKcnt], curFileName);	// save current file name
			incSTKlinenum[incSTKcnt][0] = linenum;
			incSTKlinenum[incSTKcnt][1] = linenumBase;
			++incSTKcnt;
		}
	}
	strcpy(curFileName, inclnkFileName);	// for HED blocks
	linenumBase = linenum+1;

	return 0;
}

///************************
int aAbs(int iOPS) {

	DWORD	val;

	if( mach<0 ) {
		AsmError("ABS machine not defined");
		return 1;
	}
	if( seenNAM || nextAddr || baseAddr>=0 ) {
		AsmError("ABS after code generation started");
		return 1;
	}
	if( !GetNum(&val) ) {
		AsmError("Missing ABS base address");
		return 1;
	}
	baseAddr = nextAddr = val;
	if( isAIF ) {
		if( isROM<0 ) SetBPGM(val > 060000);	// set 'is' AIF flags if not already set
		else if( isROM==0 && val<0100000 && mach!=MACH_75 ) {
			AsmError("Invalid ROM address");
			return 1;
		}
	}
	return 0;
}

///************************
int aFin(int iOPS) {

	long	len;

	NotDone = FALSE;
	// for 87 BPGM, write program len to bytes 4 & 5 of binFile
	if( pass && mach==MACH_87 && (baseAddr<0 || baseAddr>=0100000) ) {
		len = lseek(binFile, 0, SEEK_END)-512;
		lseek(binFile, 512+4, SEEK_SET);
		write(binFile, &len, 2);
		lseek(binFile, 0, SEEK_END);
	}
	// for BPGM, if absolute, write program base address to bytes 18/19 & 24/25 of binFile
	if( pass && mach!=MACH_75 && (/*baseAddr<0 ||*/ baseAddr>=0100000) ) {
		lseek(binFile, 512+18, SEEK_SET);
		write(binFile, &baseAddr, 2);
		lseek(binFile, 512+24, SEEK_SET);
		write(binFile, &baseAddr, 2);
		lseek(binFile, 0, SEEK_END);
	}
	return 0;
}

///*****************************
int aNAM(int iOPS) {

	char	label[10];
	int		i;
	DWORD	dw;
	LABEL	*pLAB;

	if( (nextAddr>0 && baseAddr>=0 && nextAddr>baseAddr) || seenNAM ) {
		AsmError("NAM after code generation started");
		return 1;
	}
	SkipWhite(pSRC);
	if( isAIF ) {
		if( isBPGM<0 ) SetBPGM(TRUE);
		else if( isBPGM==0 ) {
			AsmError("NAM in ROM");
			return 1;
		}
	}
	if( mach==MACH_85 ) {
		if( !GetLabel(label) ) {
			AsmError("Expected unquoted program NAME");
			return 1;
		}
		while( strlen(label)<6 ) strcat(label, " ");
		label[6] = 2;	// program type (BPGM)
		label[7] = 0;
		strcpy((char*)code, label);
		codelen = 7;
		for(; codelen<26; codelen++) code[codelen] = 0;
	} else if( mach==MACH_87 ) {
		dw = 512;
		if( !GetNum(&dw) ) {
			if( GetLabel(label) && ((pLAB=FindLocal(label))!=NULL || (pLAB=FindGlobal(label))!=NULL) ) dw = BNOT(pLAB->value);
		}
		if( dw>255 ) {
			AsmError("Expected BPGM # from 0-255");
			return 1;
		}
		SkipWhite(pSRC);
		if( *pSRC!=',' ) {
			AsmError("Expected ','");
			return 1;
		}
		++pSRC;	// skip comma
		SkipWhite(pSRC);
		if( !GetLabel(label) ) {
			AsmError("Expected unquoted program NAME");
			return 1;
		}
		while( strlen(label)<4 ) strcat(label, " ");
		label[4] = 0;		// two bytes of program length (set
		strcpy((char*)code, label);
		codelen = 4;
		code[codelen++] = 0;// length of program
		code[codelen++] = 0;
		code[codelen++] = 2;// program type (BPGM)
		code[codelen++] = (BYTE)dw;// BPGM #
		for(i=0; i<10 && HEDobj[i]; i++) code[codelen++] = HEDobj[i];
		for(; i<10; i++) code[codelen++] = ' ';
		for(i=18; i<26; i++) code[codelen++] = 0;
	} else {
		AsmError("NAM for unknown MODEL (85/87)");
		return 1;
	}
	seenNAM = TRUE;
	return 0;
}

///****************************
void SpitHED() {

	int		i, j, k, day=0, month=0, year=0, hour=0, minute=0, ispm=0;
	SYSTEMTIME	curtime;

	if( ListOn ) {
		lineN = 0;
		if( pageN!=0 ) {	// if listing ON and first page, spit a page WITHOUT a FF char
			if( pass ) write(lstFile, &FF, 1);
			++pageN;
		} else {
			pageN = 1;
		}
	}
	if( pass && ListOn ) {
#if defined __linux__ || defined __APPLE__
		time_t t;

		t = time(NULL);
		curtime = localtime(&t);
		year = curtime->tm_year % 100;
		month = curtime->tm_mon;
		day = curtime->tm_mday;
		hour = curtime->tm_hour % 12;
		minute = curtime->tm_min;
		ispm = curtime->tm_hour / 12;
#else
		GetLocalTime(&curtime);
		year = curtime.wYear % 100;
		month = curtime.wMonth-1;
		day = curtime.wDay;
		hour = curtime.wHour % 12;
		minute = curtime.wMinute;
		ispm = curtime.wHour / 12;
#endif

		for(i=0; i<lineW; i++) write(lstFile, "=", 1);
		write(lstFile, eolseq, strlen(eolseq));
		++lineN;

		strcpy(ScratBuf, "| src: ");
		strcat(ScratBuf, HEDsrc);
		for(i=strlen(ScratBuf); i<hedW1; i++) ScratBuf[i] = ' ';
		sprintf(ScratBuf+hedW1-11, "%02d %3.3s %02d |  ", day, Months+3*month, year);
		while( strlen(ScratBuf)<hedW1+hedW2 ) strcat(ScratBuf, " ");
		strcat(ScratBuf, "|        |");
		strcat(ScratBuf, eolseq);
        write(lstFile, ScratBuf, strlen(ScratBuf));
		++lineN;

		strcpy(ScratBuf, "| obj: ");
		strcat(ScratBuf, HEDobj);
		for(i=strlen(ScratBuf); i<hedW1; i++) ScratBuf[i] = ' ';
		sprintf(ScratBuf+hedW1-11, "%02d:%02d %2.2s  |  ", hour, minute, ampm+2*ispm);
		strcat(ScratBuf, HEDbuf);
		strcat(ScratBuf, "|        |");
		i = pageN;
		for(j=strlen(ScratBuf)-3; i; j--) {
			k = (i%10)+'0';
			ScratBuf[j] = (char)k;
			i /= 10;
		}
		strcat(ScratBuf, eolseq);
        write(lstFile, ScratBuf, strlen(ScratBuf));
		++lineN;

		for(i=0; i<lineW; i++) write(lstFile, "=", 1);
		write(lstFile, eolseq, strlen(eolseq));
		++lineN;
	} else if( ListOn ) lineN += 4;	// to keep first pass line#/page# up to date for Table of Contents
}

///***************************
int aHED(int iOPS) {

	// HED "quoted string"[, max_line_width[, max_lines_per_page]]
	// minimum max_line_width = 80
	///X <- 31 chars (include X's)-> X
	///                               X      <- 51 chars (includes X's) ->                               X
	///                                                                                                   X 10 chr X
	///=============================================================================================================
	///| src: MYROMSxxxx   01/30/84  |  1234567812345678123456781234567812345678123456781234567812345678  |        |
	///| obj: MYROM        10:33 am  |  SYSTEM TABLES                                                     |      2 |
	///=============================================================================================================
	///   filenames_datetime                HED_string                                                       page#
	///
	/// page# is always 10 chars wide
	///
	/// max filenames_datetime_width = (max_line_width-10)/2
	/// desired filenames_datetime_width = max(strlen(source_name), strlen(object_name))+20
	/// filenames_datetime_width = min(desired, max)
	///
	/// HED_string_width = max_line_width - 10 - filenames_datetime_width

	/// max len of HED "quoted string" is 64D bytes

	// need to keep track of page#

	long	i;
	long	desiredFNDTW, maxFNDTW;
	DWORD	dw;

	SkipWhite(pSRC);
	/// get HEDbuf quoted string
	if( *pSRC!='"' ) {
		AsmError("Expected quoted string");
		return 1;
	}
	++pSRC;
	if( pass==0 ) {
		if( TOCcnt+1>sizeof(TOC)/sizeof(TOC[0]) ) {
			AsmError("Too many HED pages");
			return 1;
		}
		TOC[TOCcnt].pHED = (BYTE*)pSRC;
	}
	for(i=0; i<63 && *pSRC && *pSRC!='"'; i++) {
		if( *pSRC=='\\' ) ++pSRC;
		HEDbuf[i] = *pSRC++;
	}
	HEDbuf[i] = 0;
	while( *pSRC && *pSRC!='"' ) {	// skip chars beyond first 63
		if( *pSRC=='\\' ) ++pSRC;
		if( *pSRC ) ++pSRC;
	}
	if( *pSRC=='"' ) ++pSRC;
	SkipWhite(pSRC);
	/// get optional lineW and lineH (lines/page) arguments
	if( *pSRC==',' ) {
		++pSRC;
		SkipWhite(pSRC);
		if( !GetNum(&dw) ) {
			AsmError("Expected chars/line after comma");
			return 1;
		}
		lineW = (long)dw;
		SkipWhite(pSRC);
		/// get optional lineH (lines/page) argument
		if( *pSRC==',' ) {
			++pSRC;
			SkipWhite(pSRC);
			if( !GetNum(&dw) ) {
				AsmError("Expected lines/page after comma");
				return 1;
			}
			lineH = (long)dw;
			SkipWhite(pSRC);
		} else lineH = 60;
	}// else {lineW = 95; lineH = 60;}

	/// max filenames_datetime_width = (max_line_width-10)/2
	/// desired filenames_datetime_width = max(strlen(source_name), strlen(object_name))+20
	/// filenames_datetime_width = min(desired, max)
	///
	/// HED_string_width = max_line_width - 10 - filenames_datetime_width

	hedW3 = 10;
	maxFNDTW = (lineW-hedW3)/2;
	desiredFNDTW = max(strlen(srcFileName), strlen(binFileName))+20;
	hedW1 = max(31, min(desiredFNDTW, maxFNDTW));
	hedW2 = lineW - hedW3 - hedW1;
	if( strlen(HEDbuf) > hedW2-4 ) HEDbuf[hedW2-4] = 0;	// shorten HEDbuf string if too lone for box
	while( strlen(HEDbuf) < hedW2-2 ) strcat(HEDbuf, " ");	// else lengthn HEDbuf string if too short for box

	for(i=strlen(curFileName); i; ) if( curFileName[--i]=='\\' ) break;
	if( curFileName[i]=='\\' ) ++i;
	strcpy(HEDsrc, curFileName+i);
	HEDsrc[hedW1-19] = 0;	// clip any LNK/INC'd filenames that are longer than the original source file's name

	SpitHED();
	if( pass==0 ) {
		TOC[TOCcnt++].pageN = pageN;
	}
	isHEDline = TRUE;
	return 0;
}

/*****************************************
 * Empty lines and full-line comments are output as-is.
 * Labels are output.
 * Instructions start in column startcol+9.
 * Instruction arguments start in column startcol+14.
 * End-of-line comments start in column startcol+29 (unless pushed out by argument list)
 ******************************************/
void CleanSrc(char *outbuf, int startcol, int line) {

	char	*pS;
	int		col=startcol, i, op=0;

	pS = pLines[line];
	if( *pS==0 || *pS=='!' ) {	// empty line or full-line comment, just output, no formatting
		while( *pS ) {
			if( *pS=='\t' ) {
				do {
					*outbuf++ = ' ';
					++col;
				} while( ((col-startcol)&3) );	// tab width = 4
				++pS;
			} else {
				*outbuf++ = *pS++;
				++col;
			}
		}
		*outbuf = 0;
	} else {
		if( *pS!=' ' || *pS=='\t' ) {	// label, output it
			for(; *pS && *pS!=',' && *pS!=' ' && *pS!='\t'; ++col) *outbuf++ = *pS++;
		}
		for(; col<startcol+9; col++) *outbuf++ = ' ';	// blank fill to opcode column
		while( *pS==' ' || *pS=='\t' ) ++pS;	// skip any blanks before opcode (or EOL-comment or EOL)
		if( *pS && *pS!='!' ) {	// if not at EOL or EOL-comment, must have OPCODE
			if( !strnicmp(pS, "HED", 3) || !strnicmp(pS, "ASP", 3) || !strnicmp(pS, "ASC", 3) || !strnicmp(pS, "INC", 3) || !strnicmp(pS, "LNK", 3) ) op = 1;
			for(i=0; i<4 && *pS && *pS!=' ' && *pS!='\t'; i++, col++) *outbuf++ = *pS++;	// copy OPCODE
			for(; col<startcol+14; col++) *outbuf++ = ' ';
			while( *pS==' ' || *pS=='\t' ) ++pS;	// skip tabs and spaces after OPCODE
			if( *pS!='!' ) {	// if not EOL-comment (either EOL or opcode has arguments
				switch( op ) {	// any special opcode argument handling
				 case 1:	// hed
				 	while( *pS ) {
						*outbuf++ = *pS++;	// copy quoted string with no space compression
						++col;
						if( *pS=='"' ) break;	// bail at closing quote
				 	}
/// WARNING: case 1 falls through into default!!!
				 default:
					while( *pS ) {	// while opcode has arguments, compact them and output them
						if( *pS=='\'' ) {	// character, might be space, don't strip it
							*outbuf++ = *pS++;				// output ' char
							if( *pS ) *outbuf++ = *pS++;	// output char
							if( *pS ) *outbuf++ = *pS++;	// output ' char
						} else if( *pS==' ' || *pS=='\t' ) {
							if( *++pS=='!' ) break;	// skip spaces and tabs, break when reach EOL-comment
						} else {
							*outbuf++ = *pS++;
							++col;
						}
					}
					break;
				}
			}
		}
		if( *pS=='!' ) {	// if EOL-comment
			do {	// space to proper column
				*outbuf++ = ' ';
				++col;
			} while( col<startcol+28 );
			while( *pS ) *outbuf++ = *pS++;	// copy the EOL-comment
		}
		*outbuf = 0;
	}
}

/********************************************************
 * Assemble the code, breaking into lines if first pass *
 ********************************************************/
void Asm() {

	int		i, j, k, numops;
	DWORD	val;
	char	label[10], label2[10], typ, nb;
	LABEL	*pLAB;
	char	*pS;

	numops = sizeof(OPS)/sizeof(OPS[0]);
	AIFcnt = 0;	// not inside any AIFs
	isAIF = 1;	// assemble code
	isAIFdone = 0;
	baseAddr = -1;	// BPGM and/or no ABS statement seen yet
	nextAddr = 0;	// default start address
	drp = arp = dcm = -1;
	GDADcnt = GEQUcnt = LCODcnt = LDADcnt = LEQUcnt = LAIFcnt = 0;
	Okay = InCommentBlk = 0;

	if( pass ) {
		// dump titles up front, centered
		for(i=0; i<TitlesCnt; i++) {	// for each title line
			j = (lineW-strlen((char*)Titles[i]))/2;
			for(k=0; k<j; k++) ScratBuf[k] = ' ';
			ScratBuf[k] = 0;
			strcat((char*)ScratBuf, (char*)Titles[i]);
			strcat((char*)ScratBuf, eolseq);
			write(lstFile, ScratBuf, strlen((char*)ScratBuf));
		}
		// dump table of contents up front
		if( TOCcnt ) {
			sprintf((char*)ScratBuf,"%s         TABLE OF CONTENTS%s%s",eolseq,eolseq,eolseq);
			write(lstFile, ScratBuf, strlen((char*)ScratBuf));
		}
		for(i=0; i<TOCcnt; i++) {
			BYTE *pHT;

			if( secHED ) sprintf((char*)ScratBuf, "      Section % 3d  ", TOC[i].pageN);
			else sprintf((char*)ScratBuf, "         Page % 3d  ", TOC[i].pageN);
			for(pHT=TOC[i].pHED, j=strlen((char*)ScratBuf); *pHT && *pHT!='"'; ) ScratBuf[j++] = *pHT++;
			ScratBuf[j] = 0;
			strcat((char*)ScratBuf, eolseq);
			write(lstFile, ScratBuf, strlen((char*)ScratBuf));
		}
		if( TitlesCnt || TOCcnt ) write(lstFile, &FF, 1);	// page break

		// write BPGM header
		if( seenNAM ) {	// need to write Volume and Directory sectors
			memset(ScratBuf, 0, 512);	// preclear both 'sectors'
			/// set up Volume sector
			ScratBuf[0] = 0x80;
			strcpy(ScratBuf+2, "HFSLIF");
			ScratBuf[11] = 1;	// DIR start sector
			ScratBuf[12] = 16;	// LIF ID for HP 3000 series
			ScratBuf[19] = 1;	// DIR sector count
			/// set up DIR sector
			strcpy(ScratBuf+256, "          ");
			strcpy(ScratBuf+256, HEDobj);
			i = strlen(HEDobj);
			if( i<10 ) ScratBuf[256+i] = ' ';
			ScratBuf[266] = 0xE0;
			ScratBuf[267] = 0x08;	// BPGM file type
			ScratBuf[271] = 2;		// file start sect#
			val = (finLen+255)/256;	// len of file in sectors
			ScratBuf[272] = (BYTE)((val>>24) & 0x0FF);
			ScratBuf[273] = (BYTE)((val>>16) & 0x0FF);
			ScratBuf[274] = (BYTE)((val>> 8) & 0x0FF);
			ScratBuf[275] = (BYTE)((val    ) & 0x0FF);
			ScratBuf[282] = 0x80;	// always
			ScratBuf[283] = 1;		// entire file is on this volume
			ScratBuf[284] = finLen & 0x0FF;
			ScratBuf[285] = (finLen>>8) & 0x0FF;	// byte length of file
			ScratBuf[286] = 0;
			ScratBuf[287] = 1;	// bytes/logical record
			ScratBuf[298] = ScratBuf[299] = 0xFF;	// next available directory entry
			write(binFile, ScratBuf, 512);
		}
	}
	seenNAM = FALSE;
	Public = 0;

	// now process the file, line by line
	linenum = 0;
	NotDone = TRUE;
	lineN = pageN = 0;
	for(linenum=linenumBase=incSTKcnt=0; NotDone && linenum<linesCnt; linenum++) {
		pSRC = pLines[linenum];
		codelen = BSZcnt = 0;
		AIFstmt = 0;
		ErrBuf[0] = WarnBuf[0] = 0;
		isHEDline = FALSE;

		if( linenum==0 ) {
			BOOL	skipline;

			skipline = FALSE;
			/// Process the first-line !(CMD) line if it exists
			if( !strnicmp(pSRC, "!(CMD)", 6) ) {
				skipline = TRUE;
				pSRC += 6;
				SkipWhite(pSRC);
				while( *pSRC ) {
					if( !strncmp(pSRC, "-s", 2) ) {
						pSRC += 2;
						if( pass==0 ) secHED = !secHED;
					} else if( !strncmp(pSRC, "-#", 2) ) {
						pSRC += 2;
						if( pass==0 ) StripLineNumbers = FALSE;
					} else if( !strncmp(pSRC, "-n", 2) ) {
						pSRC += 2;
						NumObj = atoi(pSRC)-1;
						while( isdigit(*pSRC) ) ++pSRC;
						if( NumObj<0 || NumObj>99 ) NumObj = 2;
						NumObjDef = NumObj;
						SetFormat();
					} else if( !strncmp(pSRC, "-a", 2) ) {
						pSRC += 2;
						if( pass==0 ) showAIF = !showAIF;
					} else if( !strncmp(pSRC, "-c", 2) ) {
						pSRC += 2;
						if( pass==0 ) CleanLST = !CleanLST;
					} else if( !strncmp(pSRC, "-h", 2) ) {
						pSRC += 2;
						if( pass==0 ) HexOut ^= 1;
					} else if( !strncmp(pSRC, "-w", 2) ) {
						pSRC += 2;
						if( pass==0 ) DoWarnings = !DoWarnings;
					} else if( !strncmp(pSRC, "-o", 2) ) {
						pSRC += 2;
						SkipWhite(pSRC);
						if( *pSRC=='"' ) {	/// if quoted name
							++pSRC;
							for(i=0; i<PATHLEN-1 && *pSRC && *pSRC!='"'; i++) binFileName[i] = *pSRC++;
						} else {
							for(i=0; i<PATHLEN-1 && *pSRC && *pSRC!=' ' && *pSRC!='\t'; ++i) binFileName[i] = *pSRC++;
						}
						binFileName[i] = 0;
					} else if( !strncmp(pSRC, "-85", 3) ) {
						pSRC += 3;
						if( pass==0 ) {
							if( mach<0 ) SetMach(MACH_85);
							else if( mach!=MACH_85 ) AsmWarning("Conflicting MODEL specification");
						}
					} else if( !strncmp(pSRC, "-R85", 4) || !strncmp(pSRC, "-r85", 4) ) {
						pSRC += 4;
						if( pass==0 ) {
							if( mach<0 ) {
								SetMach(MACH_85);
								SetBPGM(FALSE);
							} else if( mach!=MACH_85 ) AsmWarning("Conflicting MODEL specification");
						}
					} else if( !strncmp(pSRC, "-B85", 4) || !strncmp(pSRC, "-b85", 4) ) {
						pSRC += 4;
						if( pass==0 ) {
							if( mach<0 ) {
								SetMach(MACH_85);
								SetBPGM(TRUE);
							} else if( mach!=MACH_85 ) AsmWarning("Conflicting MODEL specification");
						}
					} else if( !strncmp(pSRC, "-87", 3) ) {
						pSRC += 3;
						if( pass==0 ) {
							if( mach<0 ) SetMach(MACH_87);
							else if( mach!=MACH_87 ) AsmWarning("Conflicting MODEL specification");
						}
					} else if( !strncmp(pSRC, "-R87", 4) || !strncmp(pSRC, "-r87", 4) ) {
						pSRC += 4;
						if( pass==0 ) {
							if( mach<0 ) {
								SetMach(MACH_87);
								SetBPGM(FALSE);
							} else if( mach!=MACH_87 ) AsmWarning("Conflicting MODEL specification");
						}
					} else if( !strncmp(pSRC, "-B87", 4) || !strncmp(pSRC, "-b87", 4) ) {
						pSRC += 4;
						if( pass==0 ) {
							if( mach<0 ) {
								SetMach(MACH_87);
								SetBPGM(TRUE);
							} else if( mach!=MACH_87 ) AsmWarning("Conflicting MODEL specification");
						}
					} else if( !strncmp(pSRC, "-75", 3) ) {
						pSRC += 3;
						if( pass==0 ) {
							if( mach<0 ) SetMach(MACH_75);
							else if( mach!=MACH_75 ) AsmWarning("Conflicting MODEL specification");
						}
					} else {
						if( pass==0 ) {
							sprintf((char*)ScratBuf, "Unknown !(CMD) argument: %s", pSRC);
							AsmError((char*)ScratBuf);
						}
					}
					SkipWhite(pSRC);
				}
			}
			if( pass==0 ) {
				/// build the .BIN and .LST file names
				if( binFileName[0]==0 ) {	/// if NO -o BIN file name specified, use the srcFileName as the template for the BIN and LST files
					strcpy(binFileName, srcFileName);
					for(i=strlen(binFileName)-1; i && binFileName[i]!='.'; i--) ;
					if( i>0 && binFileName[i]=='.' ) binFileName[i] = 0;
					strcpy(lstFileName, binFileName);
					strcpy(pubFileName, binFileName);
					strcat(binFileName, ".bin");
				} else {					/// else use the specified BIN file name for the BIN and LST files
					strcpy(lstFileName, binFileName);
					strcpy(pubFileName, binFileName);
					for(i=strlen(lstFileName)-1; i && lstFileName[i]!='.'; i--) ;
					if( i>0 && lstFileName[i]=='.' ) {
						lstFileName[i] = 0;
						pubFileName[i] = 0;
					}
				}
				strcat(lstFileName, ".lst");
				strcat(pubFileName, ".pub");

				/// set up binFileName in HEDobj for SpitHED() and aNAM()
				for(i=strlen(binFileName); i; ) if( binFileName[--i]=='\\' ) break;
				if( binFileName[i]=='\\' ) ++i;
				strcpy(HEDobj, binFileName+i);

				/// CREATE LIST FILE
				if( -1==(lstFile=open(lstFileName, _O_BINARY | _O_RDWR | _O_CREAT | _O_TRUNC, _S_IREAD | _S_IWRITE)) ) {
					printf("Error creating .LST file %s", lstFileName);
					exit(-1);
				}

				/// CREATE BIN FILE
				if( -1==(binFile=open(binFileName, _O_BINARY | _O_RDWR | _O_CREAT | _O_TRUNC, _S_IREAD | _S_IWRITE)) ) {
					printf("Error creating .BIN file %s", binFileName);
					exit(-1);
				}
			}
			if( skipline ) continue;	// don't list the !(CMD) line
		}
		/// auto-strip leading line numbers
		if( StripLineNumbers && isdigit(*pSRC) ) {
			while( isdigit(*pSRC) ) ++pSRC;	// skip all leading digits
			if( *pSRC==' ' ) ++pSRC;	// skip ONE space after line#s
		}
		/// process any !() meta commands
		if( !strnicmp(pSRC, "!(COM+)", 7) ) ++InCommentBlk;	// increasing
		else if( !strnicmp(pSRC, "!(COM-)", 7) ) {
			if( InCommentBlk ) --InCommentBlk;	// decreasing
			else {
				AsmError("No matching COM+");
				continue;
			}
		}
		if( !InCommentBlk ) {	// skip all of this until all comment blocks are finished
			if( *pSRC=='!' ) {	// if full-line comment
				if( *++pSRC=='(' && isAIF ) {	// might have some meta-commands to process
					if( !strnicmp(pSRC, "(OK)", 4) ) Okay = 1;	// no WARNINGs until after the next line of actual code is assembled
					else if( !strnicmp(pSRC, "(PUBLIC)", 8) ) {
						Okay = 1;
						if( pass==0 ) Public = 1;
					} else if( !strnicmp(pSRC, "(UNREF+)", 8) ) {
						if( UnRef ) {
							AsmWarning("UNREF was already on");
						}
						UnRef = 1;
					} else if( !strnicmp(pSRC, "(UNREF-)", 8) ) {
						if( !UnRef ) {
							AsmWarning("UNREF was already off");
						}
						UnRef = 0;
					} else if( !strnicmp(pSRC, "(TITLE)", 7) ) {	// if title line
						pSRC += 7;	// skip "(TITLE)"
						SkipWhite(pSRC);
						if( pass==0 ) {
							if( TitlesCnt >= sizeof(Titles)/sizeof(Titles[0]) ) {
								AsmError("Too many titles");
								continue;
							}
							Titles[TitlesCnt++] = (BYTE*)pSRC;
						}
						continue;	// throw this line away so far as listing, etc.
					} else if( !strnicmp(pSRC, "(#KEEP)", 7) ) StripLineNumbers = FALSE;
					else if( !strnicmp(pSRC, "(#STRIP)", 8) ) StripLineNumbers = TRUE;
					else if( !strnicmp(pSRC, "(#OBJ)", 6) ) {
						pSRC += 6;
						while( *pSRC==' ' || *pSRC=='\t' ) ++pSRC;
						k = atoi(pSRC);
						if( k==0 ) NumObj = NumObjDef;
						else if( k>0 && k<100 ) NumObj = k-1;
						else NumObj = 2;
						SetFormat();
					}
				}
			} else {
				// if first char of line is LABEL, get it
				if( GetLabel(label) ) {
					if( !isAIF ) {
						goto spitline;
					}
					if( strlen(label)>8 ) {
						AsmError("Label too long");

						continue;
					}

					// SkipWhite() has already been done in GetLabel()
					nb = HexOut ? BASEH : BASEO;	// default numbase
					if( label[0]=='@' ) {
						ATlabels[nextAddr/8] |= (1<<(nextAddr & 7));	// set flag that there's an @ label here
						drp = arp = dcm = -1;	// code label, invalidate known drp & arp
						goto asmrest;
					} else if( !strnicmp(pSRC, "DAD", 3) ) {
						typ = DAD;
locdadequ:
						pSRC += 3;
						SkipWhite(pSRC);
						if( !GetNum(&val) ) {
							if( GetLabel(label2) && ((pLAB=FindLocal(label2))!=NULL || (pLAB=FindGlobal(label2))!=NULL) ) {	// local or global label
								if( pass ) {
									++pLAB->refd;
									val = BNOT(pLAB->value);
								}
							} else {
								AsmError("Expected value after label DAD/EQU");
								continue;
							}
						}
						nb = toupper(*(pSRC-1));
						nb = (nb=='H')?BASEH:((nb=='C')?BASEC:((nb=='D')?BASED:((nb=='B')?BASEB:BASEO)));
					} else if( !strnicmp(pSRC, "EQU", 3) ) {
						typ = EQU;
						goto locdadequ;
					} else {
						drp = arp = dcm = -1;	// code label, invalidate known drp & arp
						typ = LOC;
						val = nextAddr;
					}
					// if first pass, add label to Locals[]
					if( pass==0 ) {
						Locals[locCnt].type = typ;
						Locals[locCnt].refd = 0;
						Locals[locCnt].unref = UnRef;
						Locals[locCnt].line = linenum;
						Locals[locCnt].public = Public;
						if( Public ) {
							pS = pSRC;
							while( *pS && *pS!='!' ) ++pS;
							Locals[locCnt].comment = pS;
							PUBcnt[(int)typ] = PUBcnt[(int)typ]+1;
							Public = 0;
						} else
							Locals[locCnt].comment = NULL;
						strcpy(Locals[locCnt].label, label);
						Locals[locCnt].value = val;
						Locals[locCnt].numbase = nb;
						if( typ==DAD ) ++LDADcnt;
						else if( typ==EQU ) ++LEQUcnt;
						else ++LCODcnt;
						++locCnt;
					} else {	// second pass, reset value (in case used for AIF/ELS/EIF
						if( (pLAB=FindLocal(label))!=NULL ) {
							if( linenum == pLAB->line ) {	// if on a different line than first defined
								pLAB->value = val;
								if( pLAB->type==DAD ) ++LDADcnt;
								else if( pLAB->type==EQU ) ++LEQUcnt;
								else if( pLAB->type==LOC ) ++LCODcnt;
								else AsmError("Something wrong. Label type unrecognized on second pass!");
							} else {
								AsmError("Duplicate label");
							}
						} else {
							AsmError("Something wrong. Label not found on second pass!");
						}
					}
				}
asmrest:
				// assemble rest of line (if anything after label)
				SkipWhite(pSRC);
				if( !isAIF && strnicmp(pSRC, "AIF", 3) && strnicmp(pSRC, "EIF", 3) && strnicmp(pSRC, "ELS", 3) && strnicmp(pSRC, "ELIF", 4) ) {
					goto spitline;
				}
				if( !strnicmp(pSRC, EINC+1, strlen(EINC)-3) ) {	// EINC+1 to skip the leading SPACE, strlen()-3 for the thrown away SPACE and the terminating CR/LF
					if( !incSTKcnt ) AsmError("EINC encountered without INC");
					else {
						while( *pSRC ) ++pSRC;	// skip to end-of-line
						--incSTKcnt;
						linenumBase = incSTKlinenum[incSTKcnt][1];	// recover prior linenumBase
						linenumBase += (linenum-incSTKlinenum[incSTKcnt][0]+1);	// adjust base for number of lines in INC file
						strcpy(curFileName, incSTKname[incSTKcnt]);	// restore prior filename for HED
						--lineN;	// to counteract the ++lineN at the end
						goto chkerrs;	// spit any pending errors and keep going
					}
				}
	/*
	if( nextAddr==060510 ) {
		val = 0;
	}
	*/
				if( isalpha(*pSRC) ) {	// must be opcode
					val = toupper(*pSRC);
					if( *++pSRC ) val += (((DWORD)toupper(*pSRC))<<8);
					if( *++pSRC ) val += (((DWORD)toupper(*pSRC))<<16);
					++pSRC;
					if( *pSRC==0 || *pSRC=='\t' ) {
						val += (((DWORD)' ')<<24);
						if( *pSRC ) ++pSRC;
					} else if( *pSRC ) {
						val += (((DWORD)toupper(*pSRC))<<24);
						++pSRC;
					}

					for(i=0; i<numops; i++) {
						if( val==OPS[i].opname ) {
							if( OPS[i].aProc(i) ) {
								// error occurred, so dump line to lst, but otherwise bail on this line
								while( *pSRC ) ++pSRC;	// skip rest of line
							}
							if( Okay ) --Okay;	// for WARNINGs
							break;
						}
					}
					if( i>=numops ) {
						AsmError("Unknown opcode");

						while( *pSRC ) ++pSRC;	// skip rest of line
					}
				}
				CheckEOLcomment();
			}
		}
		// line has been finished
spitline:
		if( codelen && mach<0 ) {
			AsmError("MODEL not specified before code generation started");
		}
		if( pass ) {	// push out code and listing on second pass
			if( codelen ) write(binFile, code, codelen);	// write out the code
			if( BSZcnt ) for(i=j=0; i<BSZcnt; i++) write(binFile, &j, 1);
		}
		if( ListOn && (showAIF || (isAIF && !AIFstmt)) && !isHEDline ) {
			long	ci=nextAddr;

			// now write out the listing
			sprintf(ScratBuf, OutForm.adrForm, (unsigned int)ci);
			for(i=0; i<min(OutForm.objNum, codelen); i++) {
				sprintf(ScratBuf+strlen(ScratBuf), OutForm.objForm, code[i]);
				++ci;
			}
			j = strlen(ScratBuf);
			if( OutForm.objW-j>0 ) strncpy(ScratBuf+j, Spaces, OutForm.objW-j);	// SPACE-fill to LABEL column
			ScratBuf[OutForm.objW] = 0;
			if( CleanLST ) CleanSrc(ScratBuf+strlen(ScratBuf), OutForm.objW, linenum);
			else strcat(ScratBuf, pLines[linenum]);	// copy out source code on first line of object code
			strcat(ScratBuf, eolseq);
			if( pass ) write(lstFile, ScratBuf, strlen(ScratBuf));
			++lineN;
			if( pageN && lineN>=lineH && !secHED ) SpitHED();
			while( i<codelen ) {	// need to dump more lines of object code bytes, but no source code
				sprintf(ScratBuf, OutForm.adrForm, (unsigned int)ci);
				for(j=0; j<OutForm.objNum && j+i<codelen; j++) {
					sprintf(ScratBuf+strlen(ScratBuf), OutForm.objForm, code[i+j]);
					++ci;
				}
				i += j;
				strcat(ScratBuf, eolseq);
				if( pass ) write(lstFile, ScratBuf, strlen(ScratBuf));
				++lineN;
				if( pageN && lineN>=lineH && !secHED ) SpitHED();
			}
		}
		nextAddr += codelen+BSZcnt;
chkerrs:
		if( ErrBuf[0] ) {
			printf("%s%s", ErrBuf, eolseq);
			if( pass ) {
				write(lstFile, ErrBuf, strlen(ErrBuf));
				write(lstFile, eolseq, strlen(eolseq));
			}
			++lineN;
			if( pageN && lineN>=lineH && !secHED ) SpitHED();
		}
		if( WarnBuf[0] ) {
			printf("%s%s", WarnBuf, eolseq);
			if( pass ) {
				write(lstFile, WarnBuf, strlen(WarnBuf));
				write(lstFile, eolseq, strlen(eolseq));
			}
			++lineN;
			if( pageN && lineN>=lineH && !secHED ) SpitHED();
		}
	}
	if( NotDone ) {
		AsmError("missing FIN");
		printf("%s%s", ErrBuf, eolseq);
		if( pass ) {
			write(lstFile, ErrBuf, strlen(ErrBuf));
			write(lstFile, eolseq, strlen(eolseq));
		}
		++lineN;
	}
	if( HEDbuf[0] ) {	// if any HED has been done
		pSRC = "\"LABEL REFERENCE\"";
		aHED(0);
	}
	if( pass==0 ) {
		finLen = (baseAddr==-1) ? nextAddr : (nextAddr-baseAddr);
	} else {
		if( seenNAM ) {	// if BPGM, fill out final 'sector' with 0xFF's
			ScratBuf[0]=0xFF;
			while( finLen & 0x00FF ) {
				write(binFile, ScratBuf, 1);
				++finLen;
			}
		}
	}
}

char	*HexForm="%3.3s  %4XH Refs:%d";
char	*BcdForm="%3.3s  %4XC Refs:%d";
char	*DecForm="%3.3s %5dD Refs:%d";
char	*OctForm="%3.3s %6o Refs:%d";
///***********************************************************************************
void DumpTable(LABEL *pL, long tabCnt, long typ, long firstL, long firstV, char *headstr) {

	long	iL, iV, i;

	write(lstFile, eolseq, strlen(eolseq));
	strcpy(ErrBuf, headstr);
	while( strlen(ErrBuf)<30 ) strcat(ErrBuf, " ");
	strcat(ErrBuf, headstr);
	while( strlen(ErrBuf)<60 ) strcat(ErrBuf, " ");
	strcat(ErrBuf, headstr);
	strcat(ErrBuf, eolseq);
	strcat(ErrBuf, eolseq);

	write(lstFile, ErrBuf, strlen(ErrBuf));
	iL = firstL;
	iV = firstV;
	for(i=0; i<tabCnt; i++) {
		ErrBuf[0] = 0;
		if( pL[i].refd && pL[i].type==typ ) {
			sprintf(ErrBuf, "%s",pL[i].label);
			while( strlen(ErrBuf)<9 ) strcat(ErrBuf, " ");
			if( pL[i].numbase==BASEH ) sprintf(ErrBuf+strlen(ErrBuf), HexForm, LabelType+3*pL[i ].type, (unsigned int)pL[i ].value, pL[i ].refd);
			else if( pL[i].numbase==BASEC ) sprintf(ErrBuf+strlen(ErrBuf), BcdForm, LabelType+3*pL[i ].type, (unsigned int)pL[i ].value, pL[i ].refd);
			else if( pL[i].numbase==BASED ) sprintf(ErrBuf+strlen(ErrBuf), DecForm, LabelType+3*pL[i ].type, (unsigned int)pL[i ].value, pL[i ].refd);
			else sprintf(ErrBuf+strlen(ErrBuf), OctForm, LabelType+3*pL[i ].type, (unsigned int)pL[i ].value, pL[i ].refd);

			while( strlen(ErrBuf)<30 ) strcat(ErrBuf, " ");
			while( !pL[iL].refd || pL[iL].type!=typ ) iL = pL[iL].Lnext;	// skip until we find a DAD that was refd
			sprintf(ErrBuf+strlen(ErrBuf), "%s",pL[iL].label);
			while( strlen(ErrBuf)<39 ) strcat(ErrBuf, " ");
			if( pL[i].numbase==BASEH ) sprintf(ErrBuf+strlen(ErrBuf), HexForm, LabelType+3*pL[i ].type, (unsigned int)pL[i ].value, pL[i ].refd);
			else if( pL[i].numbase==BASEC ) sprintf(ErrBuf+strlen(ErrBuf), BcdForm, LabelType+3*pL[i ].type, (unsigned int)pL[i ].value, pL[i ].refd);
			else if( pL[i].numbase==BASED ) sprintf(ErrBuf+strlen(ErrBuf), DecForm, LabelType+3*pL[i ].type, (unsigned int)pL[i ].value, pL[i ].refd);
			else sprintf(ErrBuf+strlen(ErrBuf), OctForm, LabelType+3*pL[i ].type, (unsigned int)pL[i ].value, pL[i ].refd);
//			if( pL[iL].numbase==BASEC ) sprintf(ErrBuf+strlen(ErrBuf), "%3.3s  %4XC Refs:%d", LabelType+3*pL[iL].type, (unsigned int)pL[iL].value, pL[iL].refd);
//			else if( pL[iL].numbase==BASED ) sprintf(ErrBuf+strlen(ErrBuf), "%3.3s %5dD Refs:%d", LabelType+3*pL[iL].type, (unsigned int)pL[iL].value, pL[iL].refd);
//			else sprintf(ErrBuf+strlen(ErrBuf), "%3.3s %6o Refs:%d", LabelType+3*pL[iL].type, (unsigned int)pL[iL].value, pL[iL].refd);
			iL = pL[iL].Lnext;	// skip to next one for next line

			while( strlen(ErrBuf)<60 ) strcat(ErrBuf, " ");
			while( !pL[iV].refd || pL[iV].type!=typ ) iV = pL[iV].Vnext;	// skip until we find a DAD that was refd
			sprintf(ErrBuf+strlen(ErrBuf), "%s",pL[iV].label);
			while( strlen(ErrBuf)<69 ) strcat(ErrBuf, " ");
			if( pL[i].numbase==BASEH ) sprintf(ErrBuf+strlen(ErrBuf), HexForm, LabelType+3*pL[i ].type, (unsigned int)pL[i ].value, pL[i ].refd);
			else if( pL[i].numbase==BASEC ) sprintf(ErrBuf+strlen(ErrBuf), BcdForm, LabelType+3*pL[i ].type, (unsigned int)pL[i ].value, pL[i ].refd);
			else if( pL[i].numbase==BASED ) sprintf(ErrBuf+strlen(ErrBuf), DecForm, LabelType+3*pL[i ].type, (unsigned int)pL[i ].value, pL[i ].refd);
			else sprintf(ErrBuf+strlen(ErrBuf), OctForm, LabelType+3*pL[i ].type, (unsigned int)pL[i ].value, pL[i ].refd);
//			if( pL[iV].numbase==BASEC ) sprintf(ErrBuf+strlen(ErrBuf), "%3.3s  %4XC Refs:%d", LabelType+3*pL[iV].type, (unsigned int)pL[iV].value, pL[iV].refd);
//			else if( pL[iV].numbase==BASED ) sprintf(ErrBuf+strlen(ErrBuf), "%3.3s %5dD Refs:%d", LabelType+3*pL[iV].type, (unsigned int)pL[iV].value, pL[iV].refd);
//			else sprintf(ErrBuf+strlen(ErrBuf), "%3.3s %6o Refs:%d", LabelType+3*pL[iV].type, (unsigned int)pL[iV].value, pL[iV].refd);
			iV = pL[iV].Vnext;	// skip to next one for next line
		}
		if( ErrBuf[0] ) {
			strcat(ErrBuf, eolseq);
			write(lstFile, ErrBuf, strlen(ErrBuf));
		}
	}
}

void WritePublicHeader(char *text)
{
	sprintf((char*)ScratBuf, "!%s! %s%s", eolseq, text, eolseq);
       	write(pubFile, ScratBuf, strlen(ScratBuf));
}

void WritePublicSymbol(LABEL *symbol)
{
	char	*pT = PublicType+3*(symbol->type);
	char	*label = symbol->label;
	unsigned int val = symbol->value;
	char	*comment = symbol->comment;

	sprintf((char*)ScratBuf, "%s\t%3.3s %6.6o", label, pT, val);
	if( *comment ) {
		strcat(ScratBuf, "\t");
		strcat(ScratBuf, comment);
	}
	strcat(ScratBuf, eolseq);
        write(pubFile, ScratBuf, strlen(ScratBuf));
}

void CreatePublicFile()
{
	int i;

	/// CREATE PUBLIC FILE
	if( -1==(pubFile=open(pubFileName, _O_BINARY | _O_RDWR | _O_CREAT | _O_TRUNC, _S_IREAD | _S_IWRITE)) ) {
		printf("Error creating .PUB file %s", pubFileName);
		exit(-1);
	}
	WritePublicHeader("THIS IS A GENERATED FILE - DO NOT EDIT");
	if( PUBcnt[1] ) {	// EQUs
		WritePublicHeader("Constants");
		for(i=0; i<locCnt; i++)
			if( Locals[i].public && Locals[i].type==EQU )
				WritePublicSymbol(&Locals[i]);
	}
	if( baseAddr<0 ) {	// BINARY
		if( PUBcnt[0] ) {	// DADs
			WritePublicHeader("Absolute addresses");
			for(i=0; i<locCnt; i++)
				if( Locals[i].public && Locals[i].type==DAD )
					WritePublicSymbol(&Locals[i]);
		}
		if( PUBcnt[2] ) {	// LABELs
			WritePublicHeader("Adresses relative to BINTAB");
			for(i=0; i<locCnt; i++)
				if( Locals[i].public && Locals[i].type==LOC )
				WritePublicSymbol(&Locals[i]);
		}
	} else if( PUBcnt[0] || PUBcnt[2] ) { // ROM DADs and LABELs
		WritePublicHeader("Absolute Addresses");
		for(i=0; i<locCnt; i++)
			if( Locals[i].public && Locals[i].type!=EQU )
				WritePublicSymbol(&Locals[i]);
	}
	close(pubFile);
}

/****************************************
 * SYNTAX: ASM85 [options] <input_name>	*
 ****************************************/
int main(int argc, char **argv) {

	int     		i, j, bpgm=-1;
	int				romlen;
	unsigned int	s, t;

#if defined __linux__ || defined __APPLE__
	gloFileName[0] = 0;
#else

	// construct PATH to globals file, ending in '\' or ':', ready to have filename appended later
	i = GetModuleFileName(NULL, gloFileName, PATHLEN-1);

	while( i && gloFileName[i]!='\\' && gloFileName[i]!=':' ) --i;
	if( gloFileName[i]=='\\' || gloFileName[i]==':' ) ++i;
	if( !i || gloFileName[i-1]!='\\' ) gloFileName[i++] = '\\';
	gloFileName[i] = 0;
#endif

	binFileName[0] = 0;
	locCnt = 0;

#if DEVTIME
if( argc==1 ) {
	strcpy(srcFileName, "gto.src");
	printf("%s\r\n", TitleRevStr);
	SetMach(MACH_85);
	SetBPGM(TRUE);
	showAIF = 1;
} else {
#endif

	// check input arguments and optionally print syntax
	for(i=1; i<argc; i++) {
		if( argv[i][0]=='-' || argv[i][0]=='/' ) {
			j = 1;
			switch( argv[i][j] ) {
			 case '7':	// 75 hopefully
			 	if( mach<0 ) {
					if( argv[i][j+1]!='5' ) goto defopt;
					SetMach(MACH_75);
			 	} else {
multmodel:
			 		printf("ERROR: multiple MODEL options %s\r\n", argv[i]);
			 	}
			 	break;
			 case 'B':
			 case 'b':
			 case 'R':
			 case 'r':
			 	if( mach>=0 ) goto multmodel;
			 	++j;
			 	if( argv[i][j]!='8' ) goto defopt;
				bpgm = (toupper(argv[i][1])=='B');
			 case '8':	// 85 or 87, hopefully
			 	if( mach>=0 ) goto multmodel;
				if( argv[i][j+1]=='5' ) SetMach(MACH_85);
				else if( argv[i][j+1]=='7' ) SetMach(MACH_87);
				else goto defopt;
				if( bpgm>=0 ) SetBPGM(bpgm);
				break;
			 case '#':	// disable auto-stripping of leading line numbers
				StripLineNumbers = FALSE;
				break;
			 case 'a':	// show CONDITIONAL code that isn't assembled
				showAIF = !showAIF;
				break;
			 case 'c':	// turn OFF the CLEAN list feature (reformat source code in LST file)
				CleanLST = !CleanLST;
				break;
			 case 'n':
				NumObj = atoi(argv[i]+2)-'1';
				if( NumObj<0 || NumObj>99 ) NumObj = 2;
				NumObjDef = NumObj;
				break;
			 case 'o':	// OUTPUT (bin) file name
			 	if( i+1 >= argc ) printf("ERROR: missing -o value");
			 	else {
					++i;	// move to next cmd line argument to get -o value
					strcpy(binFileName, argv[i]);
			 	}
				break;
			 case 's':	// SECTION hed breaks
				secHED = !secHED;
				break;
			 case 'w':	// QUIET - no warnings
			 	DoWarnings = !DoWarnings;
				break;
			 case 'h':	// HEX output
				HexOut ^= 1;
				break;
			 default:
defopt:
				printf("ERROR: unknown or invalid option: %s\r\n", argv[i]);
defnam:
		printf("\r\n\
%s - HP Capricorn CPU - Assembler\r\n\
  ASM85 [options] <source_code_filename> [options]\r\n\
    [options] =\r\n\
       Select the MODEL and GLOBALS with one of:\r\n\
            -85   for HP-85/83\r\n\
            -B85  for HP-85/83 Binary Program\r\n\
            -R85  for HP-85/83 ROM\r\n\
            -87   for HP-87/86\r\n\
            -B87  for HP-87/86 Binary Program\r\n\
            -R87  for HP-87/86 ROM\r\n\
            -75   for HP-75\r\n\r\n\
       -#   disables the stripping of leading line numbers\r\n\
       -a   show non-used conditional assembly blocks in listing\r\n\
       -c   disable CLEANING (reformatting) list file source code\r\n\
       -h   format addresses and object code as HEX in list file\r\n\
       -nX  X is decimal # of bytes of object code per list file line (1-99)\r\n\
       -o   OUTPUT (binary) file name\r\n\
            LIST file will be the same with .lst\r\n\
       -s   HEDs treated as SECTION breaks, not PAGE breaks\r\n\
       -w   WARNINGS off (don't show)\r\n\
  NOTE: code LABELS are CASE SENSITIVE!\r\n\
", TitleRevStr);
				exit(-1);
			}
		} else if( srcFileName[0]==0 ) {
			strcpy(srcFileName, argv[i]);
		} else {
			printf("ERROR: too many filenames: %s", argv[i]);
			exit(-1);
		}
	}

	// try to read the source file
#if DEVTIME
}
#endif
	if( srcFileName[0]==0 ) goto defnam;

	/// The .BIN and .LST files will get set up in Asm() on the first pass,
	/// after we've had a chance to look at the first line of the source code
	/// for source-file command line switches.
	UnRef = 1;	// list unreferenced labels
	LoadSrc(srcFileName);
	PUBcnt[0] = PUBcnt[1] = PUBcnt[2] = 0;

	SetFormat();

	//======================================================================
	// ASSEMBLE IT
	memset(ATlabels, 0, sizeof(ATlabels));	// make darn sure these flags are cleared!
	for(pass=0; pass<2; pass++) {
		nextAddr = 0;
		drp = arp = dcm = -1;
		ErrCnt = WarnCnt = 0;
		Asm();
	}
	romlen = lseek(binFile, 0, SEEK_END);	// set length of assembled code

	if( baseAddr>=0 && baseAddr<0100000 ) {	// if ROM, generate checksums
		if( mach==MACH_75 ) {
			sprintf(ErrBuf, "%d ROM filler bytes added%s", (8191-romlen), eolseq);
			printf("%s", ErrBuf);
			write(lstFile, ErrBuf, strlen(ErrBuf));

			lseek(binFile, 0, SEEK_SET);
			i = read(binFile, ScratBuf, min(8191, romlen));
			ScratBuf[8191] = 0;
			lseek(binFile, romlen, SEEK_SET);
			while( i<8191 ) {
				write(binFile, ScratBuf+8191, 1);
				++i;
			}
			// ROM bytes are in ScratBuf, 0-filled to end
			// now generate checksum in [8191]
			{
				unsigned int	chksum;

				for(i=0, chksum=0; i<8191; i++) {
					chksum += (unsigned int)((BYTE)ScratBuf[i]);
					while( chksum>255 ) {
						++chksum;
						chksum -= 256;
					}
				}
				ScratBuf[8191] = ~chksum;
			}
			write(binFile, ScratBuf+8191, 1);
		} else {
			sprintf(ErrBuf, "%d ROM filler bytes added%s", (8188-romlen), eolseq);
			printf("%s", ErrBuf);
			write(lstFile, ErrBuf, strlen(ErrBuf));

			lseek(binFile, 0, SEEK_SET);
			i = read(binFile, ScratBuf, min(8188, romlen));
			ScratBuf[8188] = ScratBuf[8189] = ScratBuf[8190] = ScratBuf[8191] = 0;
			lseek(binFile, romlen, SEEK_SET);
			while( i<8188 ) {
				write(binFile, ScratBuf+8188, 1);
				++i;
			}
			// ROM bytes are in ScratBuf, 0-filled to end
			// now generate secondary checksum in [8188] and [8189]
			{
				unsigned int	r26, r27, r45, w;
				int		j;

				for(i=t=0; i<8188; i+=2) {
					w = (unsigned int)(BYTE)ScratBuf[i] + (((unsigned int)(BYTE)ScratBuf[i+1])<<8);
					for(j=0; j<16; j++) {
						r26 = t & 0xff;
						r27 = t>>8;
						r45 = r27;
						r27 = (r27<<4) & 0xff;
						r26 = ((r26<<1) & 0xff) | (w & 1);
						w >>= 1;
						r45 ^= r26;
						r45 ^= r27;
						if( r45 & 1 ) r45 = (r45 + 0x80) & 0xff;
						t = ((t<<1) & 0xffff) | (r45 >>7);
					}
				}
				ScratBuf[8188] = (char)(t & 0xFF);
				ScratBuf[8189] = (char)(t >> 8);
			}

			// now generate primary checksum in [8190] and [8191]
			for(i=t=0; i<8190; i+=2) t += (unsigned int)(BYTE)ScratBuf[i] + (((unsigned int)(BYTE)ScratBuf[i+1])<<8);
			s = ((t>>16) + (t & 0xffff)) & 0xffff;
			ScratBuf[8190] = (char)(255-(s & 0xFF));
			ScratBuf[8191] = (char)(255-(s>>8));
			write(binFile, ScratBuf+8188, 4);
		}
	}

	if( ListOn ) {
/*struct label {
	char	type;
	char	refd;		// 0 if never referenced, 1 if referenced during the assembly (for reference table generation)
	char	label[10];	// really only use 9 max, the others keep the structure a multiple of 4 in len
	DWORD	value;
	long	Lnext, Vnext;	//  index of next [] item when sorted by Label or Value (the [] is 'sorted' by the order in which the labels were encountered)
};
typedef struct label LABEL;
long	GfirstL, GfirstV, LfirstL, LfirstV;	// index of first label in Globals[] and Labels[] when sorted by Label or Value
*/
		long	gl;		// 0 for Globals, 1 for Locals
		LABEL	*pL;	// ptr to Globals[] or to Locals[]
		long	scnt;	// # of labels sorted so far
		long	csrch;	// the current index being search for comparison
		long	psrch;	// index of previously compared item who's Lnext or Vnext pointed to the csrch item
		long	nsrch;	// the Next index to be searched
		long	first;	// temp copy of GfirstL/LfirstL/GfirstV/LfirstV
		long	lCnt;	// temp copy of gloCnt/locCnt
		long	unref;	// # of unreferenced local variables

		unref = (locCnt && Locals[0].unref && Locals[0].refd==0) ? 1 : 0;	// special case first local label
		for(gl=0; gl<2; gl++) {
			if( gl ) {
				pL = &Locals[0];
				lCnt = locCnt;
			} else {
				pL = &Globals[0];
				lCnt = gloCnt;
			}
			/// sort by LABEL
			pL->Lnext = -1;	// signify end of list on first item
			first = 0;
			for(scnt=1; scnt<lCnt; scnt++) {
				if( gl && pL[scnt].unref && !pL[scnt].refd ) ++unref;	// local not referenced, bump counter
				psrch = -1;
				csrch = first;
				while( strcmp(pL[scnt].label, pL[csrch].label)>=0 ) {	// keep going until we're LESS than the current label
					nsrch = pL[csrch].Lnext;
					if( nsrch==-1 ) {	// if reached the end of the sorted list, this new one is the highest/last sorted item (for now)
						pL[csrch].Lnext = scnt;	// point the previous last one to this one
						pL[scnt].Lnext = -1;		// mark this one as the last
						csrch = -1;						// signify we're done with this item
						break;
					}
					if( csrch==-1 ) break;
					psrch = csrch;
					csrch = nsrch;
				}
				if( csrch!=-1 ) {	// if found our spot in the list (and not added onto the end)...
					pL[scnt].Lnext = csrch;	// insert index 'scnt' in between psrch and csrch
					if( psrch!=-1 ) pL[psrch].Lnext = scnt;
					else first = scnt;
				}
			}
			// save 'first' ptrs
			if( gl ) LfirstL = first;
			else GfirstL = first;

			/// sort by VALUE
			pL->Vnext = -1;	// signify end of list on first item
			first = 0;
			for(scnt=1; scnt<lCnt; scnt++) {
				psrch = -1;
				csrch = first;
				while( pL[scnt].value >= pL[csrch].value ) {	// keep going until we're LESS than the current value
					nsrch = pL[csrch].Vnext;
					if( nsrch==-1 ) {	// if reached the end of the sorted list, this new one is the highest/last sorted item (for now)
						pL[csrch].Vnext = scnt;	// point the previous last one to this one
						pL[scnt].Vnext = -1;		// mark this one as the last
						csrch = -1;						// signify we're done with this item
						break;
					}
					if( csrch==-1 ) break;
					psrch = csrch;
					csrch = nsrch;
				}
				if( csrch!=-1 ) {	// if found our spot in the list (and not added onto the end)...
					pL[scnt].Vnext = csrch;	// insert index 'scnt' in between psrch and csrch
					if( psrch!=-1 ) pL[psrch].Vnext = scnt;
					else first = scnt;
				}
			}
			// save 'first' ptrs
			if( gl ) LfirstV = first;
			else GfirstV = first;
		}

		sprintf(ErrBuf, "%s== ORDER ENCOUNTERED ==       == SORTED BY LABEL ==         == SORTED BY VALUE ==", eolseq);
		write(lstFile, ErrBuf, strlen(ErrBuf));

		// dump symbol tables
		if( GDADcnt ) DumpTable(&Globals[0], gloCnt, DAD, GfirstL, GfirstV, "=== GLOBAL DADS ===");
		if( GEQUcnt ) DumpTable(&Globals[0], gloCnt, EQU, GfirstL, GfirstV, "=== GLOBAL EQUS ===");
		if( LCODcnt ) DumpTable(&Locals[0], locCnt, LOC, LfirstL, LfirstV, "=== LOCAL CODE LABELS ===");
		if( LDADcnt ) DumpTable(&Locals[0], locCnt, DAD, LfirstL, LfirstV, "=== LOCAL DADS ===");
		if( LEQUcnt ) DumpTable(&Locals[0], locCnt, EQU, LfirstL, LfirstV, "=== LOCAL EQUS ===");
		if( LAIFcnt ) DumpTable(&Locals[0], locCnt, AIF, LfirstL, LfirstV, "=== LOCAL AIFS ===");

		// dump list of any unreferenced local variables
		if( unref ) {
			sprintf(ErrBuf, "%s=== UNREFERENCED LOCALS ===%s%s", eolseq, eolseq, eolseq);
			write(lstFile, ErrBuf, strlen(ErrBuf));
			for(i=0; i<locCnt; i++) {
				if( Locals[i].unref && !Locals[i].refd ) {
					strcpy(ErrBuf, Locals[i].label);
					strcat(ErrBuf, eolseq);
					write(lstFile, ErrBuf, strlen(ErrBuf));
				}
			}
		}
	}
	sprintf(ErrBuf, "%s%ld errors%s", eolseq, ErrCnt, eolseq);
	printf("%s", ErrBuf);
	write(lstFile, ErrBuf, strlen(ErrBuf));
	sprintf(ErrBuf, "%ld warnings%s", WarnCnt, eolseq);
	printf("%s", ErrBuf);
	write(lstFile, ErrBuf, strlen(ErrBuf));

	if( ErrCnt || WarnCnt ) printf("===========================================>> DANGER, WILL ROBINSON!%s", eolseq);

	if( PUBcnt[0] || PUBcnt[1] || PUBcnt[2] ) CreatePublicFile();


	close(lstFile);
	close(binFile);

	exit(!!(ErrCnt + WarnCnt));
}

//***************
//* END OF FILE *
//***************
